"""
Define utilities for building engine decks.

Classes
-------
EngineDeck : the interface for an engine deck builder.

Attributes
----------
accepted_headers : dict
    The strings that are accepted as valid header names after converted to all lowercase
    with all whitespace removed, mapped to the enum EngineModelVariables.

required_variables : set
    Variables that must be present in an EngineDeck's DATA_FILE (Mach, altitude, etc.)

required_options : tuple
    Options that must be present in an EngineDeck's options attribute.

dependent_options : dict
    Options that may or may not be required based on the presence or value of other
    provided options.
"""

import math
import warnings

import numpy as np
import openmdao.api as om

from openmdao.utils.units import convert_units

from aviary.subsystems.propulsion.engine_model import EngineModel
from aviary.subsystems.propulsion.engine_scaling import EngineScaling
from aviary.subsystems.propulsion.engine_sizing import SizeEngine
from aviary.subsystems.propulsion.utils import (EngineModelVariables,
                                                convert_geopotential_altitude,
                                                default_units)
from aviary.utils.named_values import NamedValues, get_keys, get_items
from aviary.variable_info.variable_meta_data import _MetaData
from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.utils.csv_data_file import read_data_file


MACH = EngineModelVariables.MACH
ALTITUDE = EngineModelVariables.ALTITUDE
THROTTLE = EngineModelVariables.THROTTLE
HYBRID_THROTTLE = EngineModelVariables.HYBRID_THROTTLE
THRUST = EngineModelVariables.THRUST
GROSS_THRUST = EngineModelVariables.GROSS_THRUST
RAM_DRAG = EngineModelVariables.RAM_DRAG
FUEL_FLOW = EngineModelVariables.FUEL_FLOW
ELECTRIC_POWER = EngineModelVariables.ELECTRIC_POWER
NOX_RATE = EngineModelVariables.NOX_RATE
TEMPERATURE = EngineModelVariables.TEMPERATURE_ENGINE_T4
# EXIT_AREA = EngineModelVariables.EXIT_AREA

# EngineDeck assumes all aliases point to an enum, these are used internally only
aliases = {
    # whitespaces are replaced with underscores converted to lowercase before
    # comparison with keys
    MACH: ['m', 'mn', 'mach', 'mach_number'],
    ALTITUDE: ['altitude', 'alt', 'h'],
    THROTTLE: ['throttle', 'power_code', 'pc'],
    HYBRID_THROTTLE: ['hybrid_throttle', 'hpc', 'hybrid_power_code', 'electric_throttle'],
    THRUST: ['thrust', 'net_thrust'],
    GROSS_THRUST: ['gross_thrust'],
    RAM_DRAG: ['ram_drag'],
    FUEL_FLOW: ['fuel', 'fuel_flow', 'fuel_flow_rate'],
    ELECTRIC_POWER: 'electric_power',
    NOX_RATE: ['nox', 'nox_rate'],
    TEMPERATURE: ['t4', 'temp', 'temperature']
}

# these variables must be present in engine performance data
required_variables = {
    MACH,
    ALTITUDE,
    THROTTLE,
    THRUST
}

# EngineDecks internally require these options to have values. Input checks will set
# these options to default values in self.options if they are not provided
required_options = (
    Aircraft.Engine.SCALE_PERFORMANCE,
    Aircraft.Engine.IGNORE_NEGATIVE_THRUST,
    Aircraft.Engine.GEOPOTENTIAL_ALT,
    Aircraft.Engine.GENERATE_FLIGHT_IDLE,
    # TODO fuel flow scaler is required for the EngineScaling component but does not need
    #      to be defined on a per-engine basis, so it could exist only in the problem-
    #      level aviary_options without issue. Is this a propulsion_preprocessor task?
    Mission.Summary.FUEL_FLOW_SCALER
)

# options that are only required based on the value of another option
dependent_options = {
    Aircraft.Engine.GENERATE_FLIGHT_IDLE: (Aircraft.Engine.FLIGHT_IDLE_THRUST_FRACTION,
                                           Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION,
                                           Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION,)
}


class EngineDeck(EngineModel):
    """
    EngineModel that obtains performance data from a tabular input file or memory.

    Attributes
    ----------
    name : str ('engine')
        Object label.
    options : AviaryValues (<empty>)
        Inputs and options related to engine model.
    data : NamedVaues (<empty>)
        Engine performance data (optional). If provided, used instead of tabular data 
        file.

    Methods
    -------
    build_pre_mission
    build_mission
    get_val
    set_val
    update
    """

    def __init__(self, name='engine_deck', options=None, data: NamedValues = None):
        if data is not None:
            self.read_from_file = False
        else:
            self.read_from_file = True
            # TODO update default name to be based on filename

        # also calls _preprocess_inputs() as part of EngineModel __init__
        super().__init__(name, options)

        # copy of raw data read from data_file or memory, never modified or used outside
        #     EngineDeck
        self._original_data = {key: np.array([]) for key in EngineModelVariables}
        # working copy of engine performance data, is modified during data pre-processing
        self.data = {key: np.array([]) for key in EngineModelVariables}
        # gross thrust and ram drag are not used outside of EngineDeck, remove from
        #     working data
        self.data.pop(GROSS_THRUST)
        self.data.pop(RAM_DRAG)

        # number of data points in engine data
        self.model_length = 0

        self.throttle_min = 0.0
        self.throttle_max = 1.0
        self.hybrid_throttle_min = 0.0
        self.hybrid_throttle_max = 1.0

        # absolute tolerance for how far apart two points must be to be counted as unique
        self.mach_tol = 0.01
        self.alt_tol = 10.0  # ft
        self.thrust_tol = 1  # lbf

        # Create dict for variables present in engine data with associated units
        self.engine_variables = {}

        # TODO make this an option - disabling global throttle ranges is better to \
        #      prevent unintended extrapolation, but breaks missions using gasp-based
        #      engines that have uneven throttle ranges (need t4 constraint on mission
        #      to truly fix)
        self.global_throttle = True
        self.global_hybrid_throttle = True

        self._set_variable_flags()

        self._setup(data)

    def _preprocess_inputs(self):
        """
        Checks that provided options are valid and logically consistent. Raises errors 
        for non-recoverable issues, issues warnings for minor problems that are fixed at 
        runtime.

        Raises
        ------
        TypeError
            If provided options are not an instance of AviaryValues (or None).
        FileNotFoundError
            If the provided DATA_FILE cannot be found.
        """
        super()._preprocess_inputs()

        options = self.options

        # CHECK FOR REQUIRED OPTIONS
        additional_options = ()
        if self.read_from_file:
            additional_options = (Aircraft.Engine.DATA_FILE,)

        for key in additional_options + required_options:
            if key not in options:
                warnings.warn(
                    f'<{key}> is a required option for EngineDecks, but has not been '
                    f'specified for EngineDeck <{self.name}>. The default value will be '
                    'used.')

                val = _MetaData[key]['default_value']
                units = _MetaData[key]['units']
                self.set_val(key, val, units)
        # check dependent options
        for key in dependent_options:
            if self.get_val(key):
                for item in dependent_options[key]:
                    if item not in options:
                        val = _MetaData[item]['default_value']
                        units = _MetaData[item]['units']
                        self.set_val(item, val, units)

        # LOGIC CHECKS
        if self.get_val(Aircraft.Engine.GENERATE_FLIGHT_IDLE):
            idle_min = self.get_val(Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION)
            idle_max = self.get_val(Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION)
            # Allowing idle fractions to be equal, i.e. fixing flight idle conditions
            # instead of extrapolation
            if idle_min > idle_max:
                warnings.warn(
                    f'EngineDeck <{self.name}>: Minimum flight idle fraction exceeds maximum '
                    f'flight idle fraction. Values for min and max fraction will be flipped.'
                )
                self.set_val(Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION,
                             val=idle_max)
                self.set_val(Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION,
                             val=idle_min)

        # check that sufficient information on engine scaling is provided
        # default behavior is to calculate scale factor based on thrust target
        engine_mapping = get_keys(self.options)

        # check if scale factor and thrust target are user defined and check consistency
        scale_performance = self.get_val(Aircraft.Engine.SCALE_PERFORMANCE)
        scale_factor_provided = False
        thrust_provided = False
        # was scale factor originally provided? (Not defaulted)
        if Aircraft.Engine.SCALE_FACTOR in engine_mapping:
            scale_factor_provided = True
        # was scaled thrust originally provided? (Not defaulted)
        if Aircraft.Engine.SCALED_SLS_THRUST in engine_mapping:
            thrust_provided = True

        # user provided target thrust or scale factor, but performance scaling is off
        if scale_performance and (scale_factor_provided or thrust_provided):
            UserWarning(
                f'EngineDeck <{self.name}>: Scaling targets are provided, but will be '
                'ignored because performance scaling is disabled. Set '
                'aircraft:engine:SCALE_PERFORMANCE to True to enable scaling.'
            )

    def _set_variable_flags(self):
        """
        Sets flags in EngineDeck to communicate which (non-required) variables are 
        avaliable to greater propulsion module.
        """
        engine_variables = self.engine_variables

        # TODO many of these may not be required, just independents?
        self.use_thrust = THRUST in engine_variables
        self.use_fuel = FUEL_FLOW in engine_variables
        self.use_electricity = ELECTRIC_POWER in engine_variables
        self.use_hybrid_throttle = HYBRID_THROTTLE in engine_variables
        self.use_nox = NOX_RATE in engine_variables
        self.use_t4 = TEMPERATURE in engine_variables
        # self.use_exit_area = EXIT_AREA in engine_variables

    def _setup(self, data):
        """
        Read in and process engine data:

            Check data consistency.

            Convert altitudes to geometric.

            Sort and pack data.

            Determine reference thrust.

            Normalize throttles/hybrid throttles.

            Fill flight idle points.
        """
        self._read_data(data)

        # perform consistency checks on data
        self._check_data()

        # convert geopotential altitude to geometric if required
        if self.get_val(Aircraft.Engine.GEOPOTENTIAL_ALT):
            self.data[ALTITUDE] = convert_geopotential_altitude(
                self.data[ALTITUDE])

        # sort and organize data
        self._pack_data()

        # assign reference sls thrust from engine deck, perform sanity checks
        self._set_reference_thrust()

        # normalize throttle and hybrid throttle (if included) to |0-1| scale
        self._normalize_throttle()

        # extrapolate flight idle data if requested
        if self.get_val(Aircraft.Engine.GENERATE_FLIGHT_IDLE):
            self._generate_flight_idle()

    def _read_data(self, raw_data: NamedValues):
        """
        Import tabular engine data; either from memory or from a data file.

        Parameters
        ----------
        raw_data : NamedValues (optional)
            Data provided via a NamedValues object. Will be used as data source instead
            of Aircraft.Engine.DATA_FILE if self.read_from_file is False

        Raises
        ------
        ValueError
            If non-numerical data found in DATA_FILE (not including header or comments).
        """
        # custom error messages depending on data type
        if self.read_from_file:
            message = f'<{self.get_val(Aircraft.Engine.DATA_FILE)}>'
        else:
            message = f'EngineDeck <{self.name}>'

        # get data (as NamedValues object) from data file
        if self.read_from_file:
            data_file = self.get_val(Aircraft.Engine.DATA_FILE)

            # read csv file - currently not saving comments
            raw_data = read_data_file(data_file, aliases=aliases)

        else:
            # run provided data through aliases
            # create dict of what names to change, modify outside of loop
            alias_dict = {}
            for item in get_items(raw_data):
                var = item[0]
                val = item[1][0]
                units = item[1][1]
                # quick "reverse" lookup of aliases
                for key in aliases:
                    if var in aliases[key]:
                        alias_dict[var] = key
                        break

            # replace old names with aliased ones
            for name in alias_dict:
                val, units = raw_data.get_item(name)
                raw_data.delete(name)
                raw_data.set_val(alias_dict[name], val, units)

        # Loop through all variables in provided data. Track which valid variables are
        #    included with the data and save raw data for reference
        for key in get_keys(raw_data):
            val, units = raw_data.get_item(key)
            if key in aliases:
                # Convert data to expected units. Required so settings like tolerances
                # that assume units work as expected
                try:
                    val = np.array([convert_units(i, units, default_units[key])
                                   for i in val])
                except TypeError:
                    raise TypeError(f"{message}: units of '{units}' provided for "
                                    f'<{key.name}> are not compatible with expected units '
                                    f'of {default_units[key]}')

                # Engine_variables currently only used to store "valid" engine variables
                # as defined in EngineModelVariables Enum
                self.engine_variables[key] = units

            else:
                warnings.warn(
                    f'{message}: header <{key}> was not recognized, and will be skipped')

            # save all data in self._original_data, including skipped variables
            self._original_data[key] = val

        if not self.engine_variables:
            raise UserWarning(f'No valid engine variables found in data for {message}')

        # set flags using updated engine_variables
        self._set_variable_flags()

        # Copy data from original data (never modified) to working data (changed through
        #    sorting, generating missing data, etc.)
        # self.data contains all keys in EngineModelVariables except for ram drag and
        #    gross thrust
        for key in self.data:
            self.data[key] = self._original_data[key]

    def _check_data(self):
        """
        Checks for consistency of provided thrust and drag data, ensures no required
        variables are missing, fills unused variabes with a default value of zero, and
        removes negative thrusts if requested.

        Raises
        ------
        UserWarning
            If provided net thrust does not match difference between provided gross
            thrust and ram drag within tolerance.
        """
        # custom error messages depending on data type
        if self.read_from_file:
            message = f'<{self.get_val(Aircraft.Engine.DATA_FILE)}>'
        else:
            message = f'EngineDeck <{self.name}>'

        engine_variables = self.engine_variables

        # Handle ram drag, net and gross thrust and potential conflicts in value or units
        # Warn user if they provide partial info for calulated thrust
        # Not a fail state if net thrust is still provided
        # If both net thrust and components for calculated thrust both provided, a sanity
        #   check that they match is done after reading data
        if THRUST in engine_variables:
            # if thrust is present, but gross thrust or ram drag also present raise warning
            if GROSS_THRUST in engine_variables and not RAM_DRAG in engine_variables:
                warnings.warn(f'{message} contains both net and '
                              'gross thrust. Only net thrust will be used.')
            if not GROSS_THRUST in engine_variables and RAM_DRAG in engine_variables:
                warnings.warn(f'{message} contains both net thrust '
                              'and ram drag. Only net thrust will be used.')

        if RAM_DRAG in engine_variables and GROSS_THRUST in engine_variables:
            # Check that units are the same. Variables have already been checked for valid
            # units, so it is assumed they are convertable. Prioritizes thrust units
            if engine_variables[RAM_DRAG] != engine_variables[GROSS_THRUST]:
                self.data[RAM_DRAG] = convert_units(self.data,
                                                    engine_variables[RAM_DRAG],
                                                    engine_variables[GROSS_THRUST])
                engine_variables[RAM_DRAG] = engine_variables[GROSS_THRUST]

            net_thrust_calc = self._original_data[GROSS_THRUST] \
                - self._original_data[RAM_DRAG]
            # prefer using directly provided values for net thrust vs. calculating
            if THRUST in engine_variables:
                res = abs(net_thrust_calc - self._original_data[THRUST])
                if np.any(self.thrust_tol > res):
                    raise UserWarning('Provided net thrust is not equal to difference '
                                      '(within tolerance) between gross thrust and ram '
                                      f'drag in {message}')
            else:
                # store net thrust in THRUST key instead of gross thrust
                self.data[THRUST] = net_thrust_calc
                engine_variables[THRUST] = engine_variables[GROSS_THRUST]

        # remove unneeded dependent variables from engine_variables
        if RAM_DRAG in engine_variables:
            engine_variables.pop(RAM_DRAG)
        if GROSS_THRUST in engine_variables:
            engine_variables.pop(GROSS_THRUST)

        self._set_variable_flags()

        self.model_length = len(self.data[ALTITUDE])

        # check that all required variables are present in engine data
        if not required_variables.issubset(engine_variables):
            # gather all missing required variables
            missing_variables = set()
            for var in engine_variables:
                if var in required_variables:
                    missing_variables.add(var)

            # if missing_variables is not empty
            if not missing_variables:
                raise UserWarning(f'Required variables {missing_variables} are '
                                  f'missing from {message}'
                                  )

        # Set all unused variables to default value of zero
        model = self.data
        for key in model:
            if not len(model[key]):
                model[key] = np.zeros(self.model_length)

        # removes data points with negative thrust if requested
        if self.get_val(Aircraft.Engine.IGNORE_NEGATIVE_THRUST):
            keep_idx = np.where(model[THRUST] >= 0)
            for key in model:
                model[key] = model[key][keep_idx]

    def _generate_flight_idle(self):
        """
        Generate flight idle data via extrapolation from lowest points in data set,
        bound by upper and lower constraints set by user.

        Requires sorted, packed data with normalized throttles.

        Modifies unpacked data in place, updates packed data.
        """
        def _extrapolate(array):
            """
            Linearly extrapolate variable to idle thrust point.

            Parameters
            ----------
            array : numpy.ndarray
                Data used for extrapolation.

            Returns
            -------
            rvalue : float
                Extrapolated flight idle value.
            """
            y0 = array[0]
            y1 = array[1]

            if y0 == 0 and y1 == 0:
                return 0

            rvalue = (
                y0 + (y1 - y0) * thrust_extrap_term
            )

            return rvalue

        idle_thrust_fract = self.get_val(Aircraft.Engine.FLIGHT_IDLE_THRUST_FRACTION)
        idle_min_fract = self.get_val(Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION)
        idle_max_fract = self.get_val(Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION)

        packed_data = self.packed_data

        # stored information about packed data
        mach_max_count = self.mach_max_count
        alt_max_count = self.alt_max_count
        data_indices = self.data_indices

        # Throttle is already normalized from 0 to 1. Set flight idle to -0.1, which will
        # get re-normalized to 0
        # -0.1 is chosen to avoid stretching out the data range while at the same time
        # avoiding "discontinuities" in engine data from arbitrarily small negative
        # throttle (e.g. -1e-6). Basically, this is an arbitrary number
        throttle_idle = -0.1
        hybrid_throttle_idle = 0

        idle_points = {key: np.empty(0) for key in packed_data}

        # Normally, only one idle point is needed - however, when hybrid throttle is
        # present, there needs to be a sweep of points for a given Mach/alt/throttle
        # to satisfy the interpolator's requirements for at least 3 points per dimension
        # The data values at each point in the sweep are kept identical (e.g. same thrust,
        # fuel flow, etc. as calculated by extrapolation)
        num_points = 1
        if self.use_hybrid_throttle:
            num_points = 3
            # how far apart the "fake" points should be from the actual idle point
            # this time, we want an arbitrarily small number
            h_tol = 1e-4

        for M in range(mach_max_count):
            for A in range(alt_max_count):
                # if no data at this Mach, alt index combination, skip
                if data_indices[M, A] == 0:
                    continue

                # don't generate flight idle points if thrust is already zero or negative
                # at lowest index
                if packed_data[THRUST][M, A, 0] <= self.thrust_tol:
                    continue

                # define known data for idle point (independent variables)
                idle_points[MACH] = np.append(
                    idle_points[MACH], [packed_data[MACH][M, A, 0]] * num_points)
                idle_points[ALTITUDE] = np.append(
                    idle_points[ALTITUDE], [packed_data[ALTITUDE][M, A, 0]] * num_points)
                idle_points[THROTTLE] = np.append(
                    idle_points[THROTTLE], [throttle_idle] * num_points)
                if self.use_hybrid_throttle:
                    hybrid_throttle_range = np.linspace(hybrid_throttle_idle-h_tol,
                                                        hybrid_throttle_idle+h_tol,
                                                        num_points)
                    idle_points[HYBRID_THROTTLE] = np.append(
                        idle_points[HYBRID_THROTTLE], hybrid_throttle_range)
                else:
                    idle_points[HYBRID_THROTTLE] = np.append(
                        idle_points[HYBRID_THROTTLE], hybrid_throttle_idle)

                # if there is only one data point at this Mach, alt combination, use
                # thrust fraction instead of extrapolation
                # TODO idle currently calculated using lowest index data points - this is not
                #      guaranteed to be at hybrid throttle idle point, could be negative
                if data_indices[M, A] == 1:
                    for key in packed_data:
                        if key not in [
                                MACH,
                                ALTITUDE,
                                THROTTLE,
                                HYBRID_THROTTLE,
                                THRUST]:
                            idle_value = packed_data[key][M, A, 0] * idle_thrust_fract
                            var_min = packed_data[key][M, A, -1] * idle_min_fract
                            var_max = packed_data[key][M, A, -1] * idle_max_fract

                            if idle_value < var_min:
                                idle_value = var_min
                            elif idle_value > var_max:
                                idle_value = var_max

                            idle_points[key] = np.append(idle_points[key],
                                                         [idle_value] * num_points)
                            # add Mach, alt combination to idle_points with idle power
                            # codes

                    # thrust does not get idle_min/max checks
                    idle_points[THRUST] = np.append(idle_points[THRUST],
                                                    [[packed_data[THRUST][M, A, 0]
                                                     * idle_thrust_fract]] * num_points)
                    # move to next data point
                    continue

                # calculate idle thrust as a percentage of max thrust at Mach, alt point
                idle_thrust = packed_data[THRUST][M, A, data_indices[M, A] - 1]\
                    * idle_thrust_fract

                # add this thrust point to idle_points
                idle_points[THRUST] = np.append(idle_points[THRUST],
                                                [idle_thrust] * num_points)

                # calculate thrust term for linear extrapolation
                thrust_extrap_term = (idle_thrust - packed_data[THRUST][M, A, 0]) / (
                    packed_data[THRUST][M, A, 1] - packed_data[THRUST][M, A, 0])

                # compute idle data
                for key in packed_data:
                    # skip independent variables or thrust, which is already calculated
                    if key not in [
                            MACH,
                            ALTITUDE,
                            THROTTLE,
                            HYBRID_THROTTLE,
                            THRUST]:
                        # extrapolate to idle from lowest two throttle points in data
                        idle_value = _extrapolate(packed_data[key][M, A])

                        # idle cannot be below or above user-set limits
                        var_min = packed_data[key][M, A, -1] * idle_min_fract
                        var_max = packed_data[key][M, A, -1] * idle_max_fract

                        if idle_value < var_min:
                            idle_value = var_min
                        elif idle_value > var_max:
                            idle_value = var_max

                        # store newly computed idle point
                        idle_points[key] = np.append(idle_points[key],
                                                     [idle_value] * num_points)

        # add idle points to data
        for key in packed_data:
            self.data[key] = np.append(self.data[key], idle_points[key])

        # update model length
        self.model_length = len(self.data[ALTITUDE])

        # save idle points, in case they are wanted later
        self.idle_points = idle_points

        # Re-sort and re-pack data with flight idle information to keep data
        # structures consistent
        self._pack_data()

        # Re-normalize throttle since "dummy" idle values were used
        self._normalize_throttle()

    def build_pre_mission(self, aviary_inputs):
        """
        Build components to be added to pre-mission propulsion subsystem.

        Returns
        -------
            SizeEngine component specific to this EngineDeck, used for calculating engine
            scaling factors.
        """

        return SizeEngine(aviary_options=self.options)

    def build_mission(self, num_nodes, aviary_inputs):
        """
        Creates interpolator objects to be added to mission-level propulsion subsystem.
        Interpolators must be re-generated for each ODE due to potentialy different
        num_nodes in each mission segment.

        Parameters
        ----------
        num_nodes : int
            Number of nodes present in the current Dymos phase of mission analysis.

        Returns
        -------
        engine_group : openmdao.core.Group
            An OpenMDAO group containing engine data interpolators, an EngineScaling
            component, and max throttle/max hybrid_throttle generating components as
            needed for this EngineDeck.
        """
        interp_method = self.get_val(Aircraft.Engine.INTERPOLATION_METHOD)

        engine_group = om.Group()

        # interpolator object for engine data
        engine = om.MetaModelSemiStructuredComp(
            method=interp_method, extrapolate=True, vec_size=num_nodes)

        units = default_units
        for key in self.engine_variables:
            units[key] = self.engine_variables[key]

        # add inputs and outputs to interpolator
        engine.add_input(Dynamic.Mission.MACH,
                         self.data[MACH],
                         units='unitless',
                         desc='Current flight Mach number')
        engine.add_input(Dynamic.Mission.ALTITUDE,
                         self.data[ALTITUDE],
                         units=units[ALTITUDE],
                         desc='Current flight altitude')
        engine.add_input(Dynamic.Mission.THROTTLE,
                         self.data[THROTTLE],
                         units='unitless',
                         desc='Current engine throttle')
        if self.use_hybrid_throttle:
            engine.add_input(Dynamic.Mission.HYBRID_THROTTLE,
                             self.data[HYBRID_THROTTLE],
                             units='unitless',
                             desc='Current engine hybrid throttle')
        engine.add_output('thrust_net_unscaled',
                          self.data[THRUST],
                          units=units[THRUST],
                          desc='Current net thrust produced (unscaled)')
        engine.add_output('fuel_flow_rate_unscaled',
                          self.data[FUEL_FLOW],
                          units=units[FUEL_FLOW],
                          desc='Current fuel flow rate (unscaled)')
        engine.add_output('electric_power_unscaled',
                          self.data[ELECTRIC_POWER],
                          units=units[ELECTRIC_POWER],
                          desc='Current electric energy rate (unscaled)')
        engine.add_output('nox_rate_unscaled',
                          self.data[NOX_RATE],
                          units=units[NOX_RATE],
                          desc='Current NOx emission rate (unscaled)')
        # if self.use_exit_area:
        # engine.add_output('exit_area_unscaled',
        #                   self.data[EXIT_AREA],
        #                   units='ft**2',
        #                   desc='Current exit area (unscaled)')
        engine.add_output(Dynamic.Mission.TEMPERATURE_ENGINE_T4,
                          self.data[TEMPERATURE],
                          units=units[TEMPERATURE],
                          desc='Current turbine exit temperature')

        # Create copy of interpolation component that computes max thrust for current
        # flight condition
        # NOTE max thrust is assumed to occur at maximum throttle and hybrid throttle
        #      for each flight condition
        # TODO Use solver to find throttle/hybrid throttle for maximum thrust at given flight condition?
        #      Pre-solve max throttle/hybrid throttle for each flight condition, interpolate on
        #      reduced data set?
        if self.use_thrust:
            if self.global_throttle or (self.global_hybrid_throttle
                                        and self.use_hybrid_throttle):
                # create IndepVarComp to pass maximum throttle is to max thrust interpolator
                fixed_throttles = om.IndepVarComp()
                if self.global_throttle:
                    fixed_throttles.add_output('throttle_max',
                                               val=np.ones(num_nodes) *
                                               self.throttle_max,
                                               units='unitless',
                                               desc='Engine maximum throttle')
                if self.global_hybrid_throttle and self.use_hybrid_throttle:
                    fixed_throttles.add_output('hybrid_throttle_max',
                                               val=np.ones(num_nodes) *
                                               self.hybrid_throttle_max,
                                               units='unitless',
                                               desc='Engine maximum hybrid throttle')
            if not (self.global_throttle or (self.global_hybrid_throttle
                                             and self.use_hybrid_throttle)):
                interp_throttles = om.MetaModelSemiStructuredComp(method=interp_method,
                                                                  extrapolate=False,
                                                                  vec_size=num_nodes)

                packed_data = self.packed_data
                mach_table = np.array([])
                alt_table = np.array([])

                for M in range(self.mach_max_count):
                    for A in range(self.alt_max_count):
                        if self.data_indices[M, A] != 0:
                            mach_table = np.append(
                                mach_table, packed_data[MACH][M, A, 0])
                            alt_table = np.append(
                                alt_table, packed_data[ALTITUDE][M, A, 0])

                # add inputs and outputs to interpolator
                interp_throttles.add_input(Dynamic.MACH,
                                           mach_table,
                                           units='unitless',
                                           desc='Current flight Mach number')
                interp_throttles.add_input(Dynamic.ALTITUDE,
                                           alt_table,
                                           units=units[ALTITUDE],
                                           desc='Current flight altitude')
                if not self.global_throttle:
                    interp_throttles.add_output('throttle_max',
                                                self.throttle_max,
                                                units='unitless',
                                                desc='max throttle avaliable at current '
                                                'flight condition')
                if not self.global_hybrid_throttle and self.use_hybrid_throttle:
                    interp_throttles.add_output('hybrid_throttle_max',
                                                self.hybrid_throttle_max,
                                                units='unitless',
                                                desc='max hybrid throttle avaliable at '
                                                     'current flight condition')

            # Calculation of max thrust currently done with a duplicate of the engine
            # model and scaling components
            max_thrust_engine = om.MetaModelSemiStructuredComp(
                method=interp_method, extrapolate=False, vec_size=num_nodes)

            max_thrust_engine.add_input(Dynamic.Mission.MACH,
                                        self.data[MACH],
                                        units='unitless',
                                        desc='Current flight Mach number')
            max_thrust_engine.add_input(Dynamic.Mission.ALTITUDE,
                                        self.data[ALTITUDE],
                                        units=units[ALTITUDE],
                                        desc='Current flight altitude')
            # replace throttle coming from mission with max value based on flight condition
            max_thrust_engine.add_input('throttle_max',
                                        self.data[THROTTLE],
                                        units='unitless',
                                        desc='Current engine throttle')
            if self.use_hybrid_throttle:
                # replace hybrid throttle coming from mission with max value based on
                # flight condition
                max_thrust_engine.add_input('hybrid_throttle_max',
                                            self.data[HYBRID_THROTTLE],
                                            units='unitless',
                                            desc='Current engine hybrid throttle')
            max_thrust_engine.add_output('thrust_net_max_unscaled',
                                         self.data[THRUST],
                                         units=units[THRUST],
                                         desc='Current thrust produced')
        else:
            # If engine does not use thrust, a separate component for max thrust is not
            # necessary.
            # Add unscaled max thrust as output of interpolator, which will have a
            # default value of zero at every flight condition
            engine.add_output('thrust_net_max_unscaled',
                              self.data[THRUST],
                              units=units[THRUST],
                              desc='Current max net thrust produced (unscaled)')

        # add created subsystems to engine_group
        engine_group.add_subsystem('interpolation',
                                   engine,
                                   promotes_inputs=['*'],
                                   promotes_outputs=['*'])
        if self.use_thrust:
            if self.global_throttle or (self.global_hybrid_throttle
                                        and self.use_hybrid_throttle):
                engine_group.add_subsystem('fixed_max_throttles',
                                           fixed_throttles,
                                           promotes_outputs=['*'])

            if not (self.global_throttle or (self.global_hybrid_throttle
                                             and self.use_hybrid_throttle)):
                engine_group.add_subsystem('interp_max_throttles',
                                           interp_throttles,
                                           promotes_inputs=['*'],
                                           promotes_outputs=['*'])

            engine_group.add_subsystem(
                'max_thrust_interpolation',
                max_thrust_engine,
                promotes_inputs=['*'],
                promotes_outputs=['*'])

        engine_group.add_subsystem('engine_scaling',
                                   subsys=EngineScaling(num_nodes=num_nodes,
                                                        aviary_options=self.options),
                                   promotes_inputs=['*'],
                                   promotes_outputs=['*'])

        return engine_group

    def _set_reference_thrust(self):
        """
        Determine maximum sea-level static thrust produced by the engine (unscaled).

        Reference thrust can instead be directly provided in options, intended for use in
        cases where the deck does not include a SLS point or if a different point would
        make more physical sense for scaling.

        Perform consistency checks on thrust-scaling options based on new reference
        thrust.
        """
        engine_mapping = get_keys(self.options)

        if Aircraft.Engine.REFERENCE_SLS_THRUST not in engine_mapping:
            alt_tol = self.alt_tol
            mach_tol = self.mach_tol
            # NOTE This fails if there is no data point at SLS (within tolerance)
            sea_level_idx = (np.intersect1d(
                np.where(-alt_tol < self.data[ALTITUDE])[0],
                np.where(self.data[ALTITUDE] <= alt_tol)[0]))
            static_idx = (np.intersect1d(
                np.where(-mach_tol < self.data[MACH]),
                np.where(self.data[MACH] < self.mach_tol)))
            sls_idx = np.intersect1d(sea_level_idx, static_idx)

            if sls_idx.size == 0:
                raise UserWarning('Could not find sea-level static max thrust point for '
                                  f'EngineDeck <{self.name}>. Please review the data file '
                                  f'<{self.get_val(Aircraft.Engine.DATA_FILE)}> or '
                                  'manually specify Aircraft.Engine.REFERENCE_SLS_THRUST '
                                  'in EngineDeck options')

            reference_sls_thrust = max(self.data[THRUST][sls_idx])

            self.set_val(Aircraft.Engine.REFERENCE_SLS_THRUST,
                         reference_sls_thrust, units=self.engine_variables[THRUST])

        # Update SCALED_SLS_THRUST if required based on scaling information provided
        scale_performance = self.get_val(Aircraft.Engine.SCALE_PERFORMANCE)
        scale_factor_provided = False
        thrust_provided = False
        # was scale factor originally provided? (Not defaulted)
        if Aircraft.Engine.SCALE_FACTOR in engine_mapping:
            scale_factor_provided = True
        # was scaled thrust originally provided? (Not defaulted)
        if Aircraft.Engine.SCALED_SLS_THRUST in engine_mapping:
            thrust_provided = True
        ref_thrust = self.get_val(Aircraft.Engine.REFERENCE_SLS_THRUST, 'lbf')

        # logic tree when scale factor provided
        if scale_factor_provided:
            scale_factor = self.get_val(Aircraft.Engine.SCALE_FACTOR)
            # both scale factor and target thrust provided:
            if thrust_provided:
                scaled_thrust = self.get_val(Aircraft.Engine.SCALED_SLS_THRUST, 'lbf')
                if scale_performance:
                    if not math.isclose(scaled_thrust/ref_thrust, scale_factor):
                        # user wants scaling but provided conflicting inputs,
                        # cannot be resolved
                        raise AttributeError(
                            f'EngineModel <{self.name}>: Conflicting values provided for '
                            'aircraft:engine:scale_factor and '
                            'aircraft:engine:scaled_sls_thrust'
                        )
                else:
                    # engine is not scaled: just make sure scaled thrust = ref thrust
                    self.set_val(
                        Aircraft.Engine.SCALED_SLS_THRUST, ref_thrust, 'lbf')

            # scale factor provided, but not target thrust
            else:
                # calculate new scaled thrust value
                scaled_thrust = ref_thrust*scale_factor
                self.set_val(Aircraft.Engine.SCALED_SLS_THRUST, scaled_thrust, 'lbf')

        # neither scale factor nor target thrust are provided
        if not scale_factor_provided and not thrust_provided:
            if scale_performance:
                # user wants to scale, but provided no scaling info: default to
                # scale factor of 1, set scaled thrust = ref thrust
                scale_factor = 1
                self.set_val(
                    Aircraft.Engine.SCALE_FACTOR, scale_factor)
                self.set_val(
                    Aircraft.Engine.SCALED_SLS_THRUST, ref_thrust, 'lbf')
            else:
                # engine is not scaled: just make sure scaled thrust = ref thrust
                scaled_thrust = ref_thrust
                self.set_val(Aircraft.Engine.SCALED_SLS_THRUST, scaled_thrust, 'lbf')

    def _normalize_throttle(self):
        """
        Normalize throttle and hybrid throttle options. Requires packed data.

        Throttle is normalized to [0, 1], while hybrid throttle is normalized to two
        separate scales. Negative hybrid throttles are normalized to [-1, 0) and positive
        hybrid throttles to (0, 1], with the zero point representing an assumed "idle"
        condition based on the provided data.

        Normalization can be "global" (using max and min values from entire data set), or
        "local" (using the max and min values from each individual flight condition).
        """
        def _hybrid_throttle_norm(hybrid_throttle_list):
            """
            Normalize hybrid throttle to the scale:

            [-1 (minimum negative hybrid throttle)
            <-> 0 (idle hybrid throttle)
            <-> 1 (max positive hybrid throttle)]

            Negative normalized hybrid throttles only appear if negative hybrid throttle
            values are provided in engine data. Positive normalized hybrid throttle values
            only appear if positive hybrid throttle values are provided in engine data.

            Parameters
            ----------
            hybrid_throttle_list : (list, numpy.ndarray)
                Hybrid throttle data to be normalized.

            Returns
            -------
            norm_hybrid_list : numpy.ndarray
                Normalized hybrid throttle data from hybrid_throttle_list.
            """
            norm_hybrid_list = np.array(hybrid_throttle_list)
            # Split throttle into positive and negative components
            # (track index to preserve order)
            # Throttle points at zero do not need to be tracked - they are already
            # "normalized", and zero is always assumed to be in the normalization range
            # (max or min)
            hybrid_throttle_neg_idx = np.where(norm_hybrid_list < 0)
            if not hybrid_throttle_neg_idx[0].size == 0:
                hybrid_throttle_neg = norm_hybrid_list[hybrid_throttle_neg_idx]

                # normalize negative component from -1 to 0
                hybrid_throttle_neg_norm = normalize(hybrid_throttle_neg, maximum=0) - 1
                norm_hybrid_list[hybrid_throttle_neg_idx] = hybrid_throttle_neg_norm

            hybrid_throttle_pos_idx = np.where(norm_hybrid_list > 0)
            if not hybrid_throttle_pos_idx[0].size == 0:
                hybrid_throttle_pos = norm_hybrid_list[hybrid_throttle_pos_idx]

                # normalize positive component from 0 to 1
                hybrid_throttle_pos_norm = normalize(hybrid_throttle_pos, minimum=0)

                norm_hybrid_list[hybrid_throttle_pos_idx] = hybrid_throttle_pos_norm

            return norm_hybrid_list

        normalized_throttle = np.array([])
        normalized_hybrid_throttle = np.array([])
        throttle_min = np.array([])
        throttle_max = np.array([])

        # information on packed data
        packed_throttle = self.packed_data[THROTTLE]
        packed_hybrid_throttle = self.packed_data[HYBRID_THROTTLE]
        data_indices = self.data_indices

        # for each unique flight condition...
        for M in range(self.mach_max_count):
            for A in range(self.alt_max_count):
                if data_indices[M, A] == 0:
                    # skip point if there is no data
                    continue

                if not self.global_throttle:
                    throttle_list = normalize(
                        packed_throttle[M, A][:data_indices[M, A]+1])
                    # normalize throttles for this flight condition from 0 to 1
                    normalized_throttle = np.append(normalized_throttle, throttle_list)
                    throttle_min = np.append(throttle_min, min(throttle_list))
                    throttle_max = np.append(throttle_max, max(throttle_list))

                if not self.global_hybrid_throttle and self.use_hybrid_throttle:
                    # normalize hybrid throttles for this flight condition
                    hybrid_throttle_list = _hybrid_throttle_norm(
                        packed_hybrid_throttle[M, A][:data_indices[M, A]+1])
                    normalized_hybrid_throttle = np.append(
                        normalized_hybrid_throttle, hybrid_throttle_list)
                    hybrid_throttle_min = np.append(
                        hybrid_throttle_min, min(hybrid_throttle_list))
                    hybrid_throttle_max = np.append(
                        hybrid_throttle_max, max(hybrid_throttle_list))

        # store normalized throttle data
        if self.global_throttle:
            self.data[THROTTLE] = normalize(self.data[THROTTLE])
            self.throttle_min = min(self.data[THROTTLE])
            self.throttle_max = max(self.data[THROTTLE])
        else:
            self.data[THROTTLE] = normalized_throttle
            self.throttle_min = throttle_min
            self.throttle_max = throttle_max

        # store normalized hybrid throttle data
        if self.use_hybrid_throttle:
            if self.global_hybrid_throttle:
                norm_hybrid_throttle = _hybrid_throttle_norm(self.data[HYBRID_THROTTLE])

                self.hybrid_throttle_min = min(self.data[HYBRID_THROTTLE])
                self.hybrid_throttle_max = max(self.data[HYBRID_THROTTLE])
                self.data[HYBRID_THROTTLE] = norm_hybrid_throttle
            else:
                self.data[HYBRID_THROTTLE] = normalized_hybrid_throttle
                self.hybrid_throttle_min = hybrid_throttle_min
                self.hybrid_throttle_max = hybrid_throttle_max

        # repack data to keep it up to date
        self._pack_data()

    def _sort_data(self):
        """
        Sort unpacked engine data in order of mach number, altitude, throttle,
        hybrid throttle.
        """
        engine_data = self.data

        # sort engine data to ensure independent variables are always in
        # ascending order as required by metamodel interpolator

        # convert engine_data from dict to list so it can be sorted
        sorted_values = np.array([engine_data[key] for key in engine_data]).transpose()

        # Sort by mach, then altitude, then throttle, then hybrid throttle
        sorted_values = sorted_values[np.lexsort(
            [engine_data[HYBRID_THROTTLE],
             engine_data[THROTTLE],
             engine_data[ALTITUDE],
             engine_data[MACH]])]
        for idx, var in enumerate(engine_data):
            engine_data[var] = sorted_values[:, idx]

        self.data = engine_data

    def _pack_data(self):
        """
        Reorganize data from a dictionary of flat 2d arrays to a dictionary of 3d arrays
        organized by Mach, altitude, and data for each engine variable.
        Data is an array with a length equal to the number of unique data points at
        that Mach, alt point.
        """
        # method requires sorted data
        self._sort_data()
        # get updated data count
        self._count_data()

        mach_max_count = self.mach_max_count
        alt_max_count = self.alt_max_count
        data_max_count = self.data_max_count
        data_indices = self.data_indices

        packed_data = self.packed_data = {}
        idx = 0

        for key in self.data:
            packed_data[key] = np.zeros((mach_max_count, alt_max_count, data_max_count))

        for M in range(mach_max_count):

            for A in range(alt_max_count):
                if data_indices[M, A] == 0:
                    # skip point if there is no data
                    continue

                # number of data points is index+1
                for D in range(data_indices[M, A] + 1):
                    for key in self.data:
                        unpacked_data = self.data[key]
                        if idx < len(unpacked_data):
                            packed_data[key][M, A, D] = unpacked_data[idx]
                    idx += 1

    def _count_data(self):
        """
        Count unique data entries in the engine data for each Mach, altitude combination.
        Requires that data is sorted.

        Raises
        ------
        UserWarning
            If insufficient number of altitude points (<2) provided for a given Mach
            number.
        """
        mach_count = 0
        # First mach number must have at least one altitude associated with it
        alt_count = 1
        max_alt_count = 0
        # First mach number must have at least one data point associated with it
        data_count = 1
        max_data_count = 0

        # data_indices stores how many data points there are for a given Mach/alt combo
        data_indices = np.array([[]])

        curr_mach = curr_alt = np.inf

        mach_numbers = self.data[MACH]
        altitudes = self.data[ALTITUDE]

        # Loop through data. Keep track of last unique value (curr_*) to compare each new
        #   value with
        # Count number of altitudes per mach, number of data points per
        #   mach/altitude combination, compare with max_count
        for idx in range(self.model_length):
            mach_num = mach_numbers[idx]
            alt = altitudes[idx]

            if math.isclose(mach_num, curr_mach, abs_tol=self.mach_tol):

                if math.isclose(alt, curr_alt, abs_tol=self.alt_tol):
                    data_indices[mach_count - 1, alt_count - 1] = data_count
                    data_count += 1

                else:
                    # new altitude for this mach number, count it
                    curr_alt = alt
                    alt_count += 1
                    data_indices = extend_array(data_indices, [mach_count, alt_count])

                    if data_count > max_data_count:
                        max_data_count = data_count
                    # new altitude means reset data counter
                    data_count = 1
                    # count data associated with new altitude
                    data_indices[mach_count - 1, alt_count - 1] = 1

            else:
                # new Mach number
                # if there are less than two altitudes for this Mach number, quit
                if alt_count < 2 and mach_count > 0:
                    raise UserWarning('Only one altitude provided for Mach number '
                                      f'{mach_numbers[mach_count]:6.3f} in engine data file '
                                      f'<{self.get_val(Aircraft.Engine.DATA_FILE).name}>'
                                      )

                # record and count mach numbers
                curr_mach = mach_num
                mach_count += 1

                # new mach comes with new altitude, record and count it
                if alt_count > max_alt_count:
                    max_alt_count = alt_count
                # new mach means reset altitude counter
                curr_alt = alt
                alt_count = 1
                data_indices = extend_array(data_indices, [mach_count, alt_count])

                if data_count > max_data_count:
                    max_data_count = data_count
                # new mach means reset data counter
                data_count = 1
                # count data associated with new altitude
                data_indices[mach_count - 1, alt_count - 1] = 1

        self.mach_max_count = mach_count
        self.alt_max_count = max_alt_count
        self.data_max_count = max_data_count
        self.data_indices = data_indices.astype(int)


#####################
# UTILITY FUNCTIONS #
#####################
"""
Functions that do not directly use attributes of EngineDeck (do not require self) are 
located here. These functions are currently only used for EngineDecks and are not 
applicable to other EngineModels. If any of these functions become useful to other 
EngineModels besides EngineDeck, move them to propulsion utils.
"""


def normalize(base_list, maximum=None, minimum=None):
    """
    Normalize the given list from 0 to 1.
    Maximum or minimum of data range can be overwritten, otherwise range of list
    is assumed to contain full range of data that must be normalized.

    Parameters
    ----------
    base_list : (list, numpy.ndarray)
        Data that is to be normalized.
    maximum : float
        Overwritten maximum value of data that will scale to 1 when normalized.
    minimum : float
        Overwritten minimum value of data that will scale to 0 when normalized. 

    Returns
    -------
    norm_list : numpy.ndarray
        Normalized data from base_list.
    """
    if maximum is None:
        maximum = max(base_list)
    if minimum is None:
        minimum = min(base_list)

    norm_list = np.array([(x - minimum) / (maximum - minimum) for x in base_list])

    return norm_list


def extend_array(inp_array, size):
    """
    Extends input array such that it is at least as large as the target size in
    all dimensions. If input array is smaller in any dimension, extends input
    array to match target size in that dimension. Works on arrays up to 3
    dimensions. Returns copy of input array extended in required dimensions with newly
    created points set to 0.

    Parameters
    ----------
    inp_array : (list, numpy.ndarray)
        Array that needs to be checked for expansion.
    size : list
        List containing desired minimum length of inp_array along each dimension.

    Returns
    -------
    inp_array : numpy.ndarray
        The provided array extended in the desired dimensions with with newly created
        points set to 0.
    """
    # TODO may be built-in numpy functions that can replace this and support n dimensions
    dims = np.array(np.shape(inp_array))
    while size[0] > dims[0]:
        inp_array = np.concatenate(
            (inp_array, np.zeros((1, *dims[1:]))), 0)
        dims[0] += 1
    if len(dims) > 1:
        while size[1] > dims[1]:
            inp_array = np.concatenate(
                (inp_array, np.zeros((dims[0], 1, *dims[2:]))), 1)
            dims[1] += 1
        if len(dims) > 2:
            while size[2] > dims[2]:
                inp_array = np.concatenate(
                    (inp_array, np.zeros((*dims[:2], 1))), 2)
                dims[2] += 1

    return inp_array



import numpy as np
import openmdao.api as om
import scipy.constants as _units

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic


class InducedDrag(om.ExplicitComponent):
    """
    Calculates induced drag
    """

    def initialize(self):
        self.options.declare(
            "num_nodes", default=1, types=int,
            desc="Number of nodes along mission segment")
        self.options.declare(
            'gamma', default=1.4,
            desc='Ratio of specific heats for air.')
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        nn = self.options["num_nodes"]

        # Simulation inputs
        self.add_input(
            Dynamic.Mission.MACH, shape=(nn), units='unitless', desc="Mach number")
        self.add_input(
            Dynamic.Mission.LIFT, shape=(nn), units="lbf", desc="Lift magnitude")
        self.add_input(
            Dynamic.Mission.STATIC_PRESSURE, np.ones(nn), units='lbf/ft**2',
            desc='Static pressure at each evaulation point.')

        # Aero design inputs
        add_aviary_input(self, Aircraft.Wing.AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.SPAN_EFFICIENCY_FACTOR, 0.0)
        add_aviary_input(self, Aircraft.Wing.SWEEP, 0.0)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, 0.0)

        # Declare outputs
        self.add_output(
            "induced_drag_coeff", shape=(nn), units='unitless',
            desc="Coefficient of induced drag")

    def setup_partials(self):
        nn = self.options["num_nodes"]

        row_col = np.arange(nn)
        self.declare_partials(
            'induced_drag_coeff',
            [Dynamic.Mission.MACH, Dynamic.Mission.LIFT, Dynamic.Mission.STATIC_PRESSURE],
            rows=row_col, cols=row_col)

        wrt = [
            Aircraft.Wing.AREA,
            Aircraft.Wing.ASPECT_RATIO,
            Aircraft.Wing.SPAN_EFFICIENCY_FACTOR,
            Aircraft.Wing.SWEEP,
            Aircraft.Wing.TAPER_RATIO]

        self.declare_partials('induced_drag_coeff', wrt, rows=row_col, cols=np.zeros(nn))

    def compute(self, inputs, outputs):
        options = self.options
        gamma = options['gamma']
        aviary_options: AviaryValues = options['aviary_options']
        mach, lift, P, Sref, AR, span_efficiency_factor, SW25, TR = inputs.values()

        CL = 2.0 * lift / (Sref * gamma * P * mach ** 2)

        redux, _ = aviary_options.get_item(
            Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION, (False, None))

        if redux:
            # Adjustment for extreme taper ratios.
            # Reference:
            # ----------
            # [1] DeYoung, John. "Advanced Supersonic Technology Concept Study Reference
            # Characteristics," NASA Contractor Report 132374.

            span_efficiency_0 = 1.0 + 0.1 * AR * (.4226 * np.sqrt(AR) - .35 * TR - .143)
        else:
            span_efficiency_0 = 1.0

        if span_efficiency_factor <= 0.3:
            span_efficiency = span_efficiency_0 + span_efficiency_factor
        else:
            span_efficiency = span_efficiency_0 * span_efficiency_factor

        CDi = CL ** 2 / (np.pi * AR * span_efficiency)

        # If forward sweep, add Warner Robins Factor
        if (SW25.real < 0.0):
            deg_to_rad = _units.degree

            TH = (1.0 - TR) / (1.0 + TR) / AR
            tan_sw = np.tan(SW25 / deg_to_rad)
            COSA = 1.0 / np.sqrt(1.0 + (tan_sw - 3.0 * TH) ** 2)
            COSB = 1.0 / np.sqrt(1.0 + (tan_sw + TH) ** 2)
            CAYT = 0.5 * (
                (1.1 - .11 / (1.1 - mach * COSA))
                / (1.1 - .11 / (1.1 - mach * COSB)) - 1.0)**2

            CDi += CAYT * CL ** 2

        outputs['induced_drag_coeff'] = CDi

    def compute_partials(self, inputs, partials):
        options = self.options
        gamma = options['gamma']
        aviary_options: AviaryValues = options['aviary_options']
        mach, lift, P, Sref, AR, span_efficiency_factor, SW25, TR = inputs.values()
        redux, _ = aviary_options.get_item(
            Aircraft.Wing.SPAN_EFFICIENCY_REDUCTION, (False, None))

        if redux:
            sqrt_AR = np.sqrt(AR)
            span_efficiency_0 = 1.0 + 0.1 * AR * (.4226 * sqrt_AR - .35 * TR - .143)
            dse0_dAR = 0.1 * (.4226 * (sqrt_AR + 0.5 * AR / sqrt_AR) - .35 * TR - .143)
            dse0_dTR = -0.035 * AR

        else:
            span_efficiency_0 = 1.0

        if span_efficiency_factor <= 0.3:
            span_efficiency = span_efficiency_0 + span_efficiency_factor
        else:
            span_efficiency = span_efficiency_0 * span_efficiency_factor

        CL = 2.0 * lift / (Sref * gamma * P * mach ** 2)
        dCL_dL = 2.0 / (Sref * gamma * P * mach ** 2)
        dCL_dSref = -2.0 * lift / (Sref ** 2 * gamma * P * mach ** 2)
        dCL_dP = -2.0 * lift / (Sref * gamma * P ** 2 * mach ** 2)
        dCL_dmach = -4.0 * lift / (Sref * gamma * P * mach ** 3)

        CDi = CL ** 2 / (np.pi * AR * span_efficiency)
        dCDi_dCL = 2.0 * CL / (np.pi * AR * span_efficiency)
        dCDi_dAR = -CL ** 2 / (np.pi * AR ** 2 * span_efficiency)
        dCDi_dspan = -CL ** 2 / (np.pi * AR * span_efficiency ** 2)

        partials['induced_drag_coeff', Dynamic.Mission.MACH] = dCDi_dCL * dCL_dmach
        partials['induced_drag_coeff', Dynamic.Mission.LIFT] = dCDi_dCL * dCL_dL
        partials['induced_drag_coeff', Dynamic.Mission.STATIC_PRESSURE] = dCDi_dCL * dCL_dP
        partials['induced_drag_coeff', Aircraft.Wing.ASPECT_RATIO] = dCDi_dAR
        partials['induced_drag_coeff', Aircraft.Wing.SPAN_EFFICIENCY_FACTOR] = 0.0
        partials['induced_drag_coeff', Aircraft.Wing.SWEEP] = 0.0
        partials['induced_drag_coeff', Aircraft.Wing.TAPER_RATIO] = 0.0
        partials['induced_drag_coeff', Aircraft.Wing.AREA] = dCDi_dCL * dCL_dSref

        if span_efficiency_factor <= 0.3:
            partials['induced_drag_coeff', Aircraft.Wing.SPAN_EFFICIENCY_FACTOR] = \
                dCDi_dspan
            if redux:
                partials['induced_drag_coeff', Aircraft.Wing.ASPECT_RATIO] += \
                    dCDi_dspan * dse0_dAR
                partials['induced_drag_coeff', Aircraft.Wing.TAPER_RATIO] = \
                    dCDi_dspan * dse0_dTR
        else:
            partials['induced_drag_coeff', Aircraft.Wing.SPAN_EFFICIENCY_FACTOR] = \
                dCDi_dspan * span_efficiency_0
            if redux:
                partials['induced_drag_coeff', Aircraft.Wing.ASPECT_RATIO] += \
                    dCDi_dspan * dse0_dAR * span_efficiency_factor
                partials['induced_drag_coeff', Aircraft.Wing.TAPER_RATIO] = \
                    dCDi_dspan * dse0_dTR * span_efficiency_factor

        # If forward sweep, add Warner Robins Factor
        if (SW25.real < 0.0):
            deg_to_rad = _units.degree

            fact1 = 1.0 - TR
            fact2 = 1.0 / (1.0 + TR)
            TH = fact1 * fact2 / AR
            dTH_dTR = - (fact2 + fact1 * fact2 ** 2) / AR
            dTH_dAR = - fact1 * fact2 / AR ** 2

            tan_sw = np.tan(SW25 / deg_to_rad)
            dtansw_dsw = 1.0 / (deg_to_rad * np.cos(SW25 / deg_to_rad) ** 2)

            fact3 = 1.0 + (tan_sw - 3.0 * TH) ** 2
            COSA = 1.0 / np.sqrt(fact3)
            dCOSA_dtansw = -0.5 / fact3 ** 1.5 * 2.0 * (tan_sw - 3.0 * TH)
            dCOSA_dTH = 0.5 / fact3 ** 1.5 * 6.0 * (tan_sw - 3.0 * TH)

            fact4 = 1.0 + (tan_sw + TH) ** 2
            COSB = 1.0 / np.sqrt(fact4)
            dCOSB_dtansw = -0.5 / fact4 ** 1.5 * 2.0 * (tan_sw + TH)
            dCOSB_dTH = -0.5 / fact4 ** 1.5 * 2.0 * (tan_sw + TH)

            factA = 1.1 - mach * COSA
            factB = 1.1 - mach * COSB
            fact5 = 1.1 - .11 / factA
            fact6 = 1.1 - .11 / factB
            CAYT = 0.5 * (fact5 / fact6 - 1.0)**2
            dCAYT_dmach = (fact5 / fact6 - 1.0) * (
                -.11 * COSA / (fact6 * factA ** 2)
                + .11 * fact5 * COSB / (factB ** 2 * fact6 ** 2))
            dCAYT_dCOSA = (fact5 / fact6 - 1.0) * (-.11 * mach / (fact6 * factA ** 2))
            dCAYT_dCOSB = (fact5 / fact6 - 1.0) * (
                .11 * fact5 * mach / (factB ** 2 * fact6 ** 2))

            dCDi_dCAYT = CL ** 2
            dCDi_dCL = 2.0 * CAYT * CL

            partials['induced_drag_coeff', Dynamic.Mission.MACH] += \
                dCDi_dCL * dCL_dmach + dCDi_dCAYT * dCAYT_dmach
            partials['induced_drag_coeff', Dynamic.Mission.LIFT] += dCDi_dCL * dCL_dL
            partials['induced_drag_coeff', Aircraft.Wing.ASPECT_RATIO] += (
                dCDi_dCAYT * dTH_dAR
                * (dCAYT_dCOSA * dCOSA_dTH + dCAYT_dCOSB * dCOSB_dTH))
            partials['induced_drag_coeff', Aircraft.Wing.SWEEP] += (
                dCDi_dCAYT * dtansw_dsw
                * (dCAYT_dCOSA * dCOSA_dtansw + dCAYT_dCOSB * dCOSB_dtansw))
            partials['induced_drag_coeff',
                     Dynamic.Mission.STATIC_PRESSURE] += dCDi_dCL * dCL_dP
            partials['induced_drag_coeff', Aircraft.Wing.TAPER_RATIO] += (
                dCDi_dCAYT * dTH_dTR
                * (dCAYT_dCOSA * dCOSA_dTH + dCAYT_dCOSB * dCOSB_dTH))
            partials['induced_drag_coeff', Aircraft.Wing.AREA] += dCDi_dCL * dCL_dSref



from aviary.variable_info.enums import SpeedType
from aviary.subsystems.propulsion.propulsion_builder import CorePropulsionBuilder
from aviary.subsystems.geometry.geometry_builder import CoreGeometryBuilder
from aviary.subsystems.mass.mass_builder import CoreMassBuilder
from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.variable_info.variable_meta_data import _MetaData as BaseMetaData
from aviary.variable_info.enums import LegacyCode

GASP = LegacyCode.GASP


throttle_max = 1.0
throttle_climb = 0.956
throttle_cruise = 0.930
throttle_idle = 0.0

prop = CorePropulsionBuilder('core_propulsion', BaseMetaData)
mass = CoreMassBuilder('core_mass', BaseMetaData, GASP)
aero = CoreAerodynamicsBuilder('core_aerodynamics', BaseMetaData, GASP)
geom = CoreGeometryBuilder('core_geometry', BaseMetaData, GASP)

default_premission_subsystems = [prop, geom, aero, mass]
default_mission_subsystems = [aero, prop]

phase_info = {
    'groundroll': {
        'num_segments': 3,
        'throttle_setting': throttle_max,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': True,
        'clean': False,
        'initial_ref': (100., 'kn'),
        'initial_bounds': ((0., 500.), 'kn'),
        'duration_ref': (100., 'kn'),
        'duration_bounds': ((50., 1000.), 'kn'),
        'control_order': 1,
        'order': 3,
        'opt': True,
    },
    'rotation': {
        'num_segments': 2,
        'throttle_setting': throttle_max,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': True,
        'clean': False,
        'initial_ref': (1.e3, 'm'),
        'initial_bounds': ((50., 5000.), 'm'),
        'duration_ref': (1.e3, 'm'),
        'duration_bounds': ((50., 2000.), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': True,
    },
    'ascent_to_gear_retract': {
        'num_segments': 2,
        'throttle_setting': throttle_max,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': False,
        'clean': False,
        'initial_ref': (1.e3, 'm'),
        'initial_bounds': ((10., 2.e3), 'm'),
        'duration_ref': (1.e3, 'm'),
        'duration_bounds': ((500., 1.e4), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': False,
    },
    'ascent_to_flap_retract': {
        'num_segments': 2,
        'throttle_setting': throttle_max,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': False,
        'clean': False,
        'initial_ref': (1.e3, 'm'),
        'initial_bounds': ((10., 2.e5), 'm'),
        'duration_ref': (1.e3, 'm'),
        'duration_bounds': ((500., 1.e4), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': False,
    },
    'ascent': {
        'num_segments': 2,
        'throttle_setting': throttle_max,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e3, 'm'),
        'initial_bounds': ((10., 2.e5), 'm'),
        'duration_ref': (1.e3, 'm'),
        'duration_bounds': ((500., 1.e4), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': False,
    },
    'climb_at_constant_TAS': {
        'num_segments': 2,
        'throttle_setting': throttle_max,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e3, 'm'),
        'initial_bounds': ((10., 2.e5), 'm'),
        'duration_ref': (1.e3, 'm'),
        'duration_bounds': ((500., 1.e4), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': False,
    },
    'climb_at_constant_EAS': {
        'num_segments': 2,
        'throttle_setting': throttle_climb,
        'input_speed_type': SpeedType.EAS,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e5, 'm'),
        'initial_bounds': ((100., 1.e7), 'm'),
        'duration_ref': (1.e5, 'm'),
        'duration_bounds': ((1.e4, 5.e4), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': False,
        'fixed_EAS': 250.,
    },
    'climb_at_constant_EAS_to_mach': {
        'num_segments': 2,
        'throttle_setting': throttle_climb,
        'input_speed_type': SpeedType.EAS,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e5, 'm'),
        'initial_bounds': ((100., 1.e7), 'm'),
        'duration_ref': (1.e5, 'm'),
        'duration_bounds': ((10.e3, 1.e6), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': True,
        'fixed_EAS': 270.,
    },
    'climb_at_constant_mach': {
        'num_segments': 2,
        'throttle_setting': throttle_climb,
        'input_speed_type': SpeedType.MACH,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e5, 'm'),
        'initial_bounds': ((100., 1.e7), 'm'),
        'duration_ref': (1.e5, 'm'),
        'duration_bounds': ((10.e3, 1.e6), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': True,
    },
    'cruise': {
        'num_segments': 5,
        'throttle_setting': throttle_cruise,
        'input_speed_type': SpeedType.MACH,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e6, 'm'),
        'initial_bounds': ((1.e4, 2.e7), 'm'),
        'duration_ref': (5.e6, 'm'),
        'duration_bounds': ((0.1e7, 3.e7), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': False,
    },
    'descent': {
        'num_segments': 3,
        'throttle_setting': throttle_idle,
        'input_speed_type': SpeedType.TAS,
        'ground_roll': False,
        'clean': True,
        'initial_ref': (1.e6, 'm'),
        'initial_bounds': ((100., 1.e7), 'm'),
        'duration_ref': (1.e5, 'm'),
        'duration_bounds': ((1.5e5, 2.e5), 'm'),
        'control_order': 1,
        'order': 3,
        'opt': True,
    },
}



import numpy as np
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.utils.named_values import NamedValues, get_keys, get_items


class DataInterpolator(om.Group):
    '''
    Group that contains interpolators that get passed training data directly through
    openMDAO connections
    '''

    def initialize(self):
        self.options.declare('num_nodes', types=int)

        self.options.declare(
            'aviary_options', types=AviaryValues,
            default=None,
            desc='Collection of Aircraft/Mission specific options')

        self.options.declare(
            'interpolator_inputs', types=NamedValues,
            desc='NamedValues object containing data for the independent variables of '
                 'interpolation, including name, value, and units'
        )

        self.options.declare(
            'interpolator_outputs', types=dict,
            desc='Dictionary describing which variables will be avaliable to the '
                 'interpolator as training data at runtime, and their units'
        )

        self.options.declare(
            'interpolation_method',
            values=['slinear', 'lagrange2', 'lagrange3', 'akima'],
            default='slinear',
            desc='Interpolation method for metamodel'
        )

    def setup(self):
        num_nodes = self.options['num_nodes']
        input_data = self.options['interpolator_inputs']
        output_data = self.options['interpolator_outputs']
        interp_method = self.options['interpolation_method']

        # interpolator object for engine data
        engine = om.MetaModelSemiStructuredComp(
            method=interp_method, extrapolate=True,
            vec_size=num_nodes, training_data_gradients=True)

        # Calculation of max thrust currently done with a duplicate of the engine
        # model and scaling components
        max_thrust_engine = om.MetaModelSemiStructuredComp(
            method=interp_method, extrapolate=True,
            vec_size=num_nodes, training_data_gradients=True)

        # check that data in table are all vectors of the same length
        for idx, item in enumerate(get_items(input_data)):
            val = item[1][0]
            if idx != 0:
                prev_model_length = model_length
            else:
                prev_model_length = len(val)
            model_length = len(val)
            if model_length != prev_model_length:
                raise IndexError('Lengths of data provided for engine performance '
                                 'interpolation do not match.')

        # add inputs and outputs to interpolator
        for input in get_keys(input_data):
            values, units = input_data.get_item(input)
            engine.add_input(input,
                             training_data=values,
                             units=units)

            if input == 'throttle':
                input = 'throttle_max'
            if input == 'hybrid_throttle':
                input = 'hybrid_throttle_max'
            max_thrust_engine.add_input(input,
                                        training_data=values,
                                        units=units)

        for output in output_data:
            engine.add_output(output,
                              training_data=np.zeros(model_length),
                              units=output_data[output])
            if output == 'thrust_net':
                max_thrust_engine.add_output('thrust_net_max',
                                             training_data=np.zeros(model_length),
                                             units=output_data[output])

        # create IndepVarComp to pass maximum throttle is to max thrust interpolator
        # currently assuming max throttle and max hybrid throttle is always 1 at every
        #   flight condition
        fixed_throttles = om.IndepVarComp()
        fixed_throttles.add_output('throttle_max',
                                   val=np.ones(num_nodes),
                                   units='unitless',
                                   desc='Engine maximum throttle')
        if 'hybrid_throttle' in input_data:
            fixed_throttles.add_output('hybrid_throttle_max',
                                       val=np.ones(num_nodes),
                                       units='unitless',
                                       desc='Engine maximum hybrid throttle')

        # add created subsystems to engine_group
        self.add_subsystem('interpolation',
                           engine,
                           promotes_inputs=['*'],
                           promotes_outputs=['*'])

        self.add_subsystem('fixed_max_throttles',
                           fixed_throttles,
                           promotes_outputs=['*'])

        self.add_subsystem('max_thrust_interpolation',
                           max_thrust_engine,
                           promotes_inputs=['*'],
                           promotes_outputs=['*'])



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Dynamic, Mission


class TaxiFuelComponent(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        self.add_input(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL,
            val=1.0,
            units="lbm/s",
            desc="fuel flow rate",
        )
        add_aviary_input(self, Mission.Summary.GROSS_MASS, val=175400.0)

        self.add_output(
            "taxi_fuel_consumed",
            val=1.0,
            units="lbm",
            desc="taxi_fuel_consumed",
        )
        self.add_output(
            Dynamic.Mission.MASS,
            val=175000.0,
            units="lbm",
            desc="mass after taxi",
        )

        self.declare_partials(
            "taxi_fuel_consumed", [
                Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL])
        self.declare_partials(
            Dynamic.Mission.MASS, Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL)
        self.declare_partials(
            Dynamic.Mission.MASS, Mission.Summary.GROSS_MASS, val=1)

    def compute(self, inputs, outputs):
        fuelflow, takeoff_mass = inputs.values()
        dt_taxi = self.options['aviary_options'].get_val(Mission.Taxi.DURATION, 's')
        outputs["taxi_fuel_consumed"] = -fuelflow * dt_taxi
        outputs[Dynamic.Mission.MASS] = takeoff_mass - outputs["taxi_fuel_consumed"]

    def compute_partials(self, inputs, J):
        dt_taxi = self.options['aviary_options'].get_val(Mission.Taxi.DURATION, 's')

        J["taxi_fuel_consumed", Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL] = -dt_taxi

        J[Dynamic.Mission.MASS, Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL] = dt_taxi



from aviary.mission.phase_builder_base import PhaseBuilderBase
from aviary.mission.initial_guess_builders import InitialGuessState, InitialGuessTime
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Dynamic
from aviary.mission.gasp_based.ode.breguet_cruise_ode import BreguetCruiseODESolution
from aviary.variable_info.variable_meta_data import _MetaData


class CruisePhase(PhaseBuilderBase):
    """
    A phase builder for a climb phase in a mission simulation.

    This class extends the PhaseBuilderBase class, providing specific implementations for
    the climb phase of a flight mission.

    Attributes
    ----------
    Inherits all attributes from PhaseBuilderBase.

    Methods
    -------
    Inherits all methods from PhaseBuilderBase.
    Additional method overrides and new methods specific to the climb phase are included.
    """
    default_name = 'cruise_phase'
    default_ode_class = BreguetCruiseODESolution

    _meta_data_ = {}
    _initial_guesses_meta_data_ = {}

    def __init__(
        self, name=None, subsystem_options=None, user_options=None, initial_guesses=None,
        ode_class=None, transcription=None, core_subsystems=None,
        external_subsystems=None, meta_data=None
    ):
        super().__init__(
            name=name, subsystem_options=subsystem_options, user_options=user_options,
            initial_guesses=initial_guesses, ode_class=ode_class, transcription=transcription,
            core_subsystems=core_subsystems, is_analytic_phase=True,
        )

    def build_phase(self, aviary_options: AviaryValues = None):
        """
        Return a new climb phase for analysis using these constraints.

        If ode_class is None, ClimbODE is used as the default.

        Parameters
        ----------
        aviary_options : AviaryValues
            Collection of Aircraft/Mission specific options

        Returns
        -------
        dymos.Phase
        """
        phase = super().build_phase(aviary_options)

        # Custom configurations for the climb phase
        user_options = self.user_options

        mach_cruise = user_options.get_val('mach_cruise')
        alt_cruise, alt_units = user_options.get_item('alt_cruise')

        phase.add_parameter(Dynamic.Mission.ALTITUDE, opt=False,
                            val=alt_cruise, units=alt_units)
        phase.add_parameter(Dynamic.Mission.MACH, opt=False,
                            val=mach_cruise)
        phase.add_parameter("initial_distance", opt=False, val=0.0,
                            units="NM", static_target=True)
        phase.add_parameter("initial_time", opt=False, val=0.0,
                            units="s", static_target=True)

        phase.add_timeseries_output("time", units="s")

        return phase


# Adding metadata for the CruisePhase
CruisePhase._add_meta_data('alt_cruise', val=0)
CruisePhase._add_meta_data('mach_cruise', val=0)

CruisePhase._add_initial_guess_meta_data(
    InitialGuessTime(),
    desc='initial guess for initial time and duration specified as a tuple')

CruisePhase._add_initial_guess_meta_data(
    InitialGuessState('mass'),
    desc='initial guess for mass')

CruisePhase._add_initial_guess_meta_data(
    InitialGuessState('initial_distance'),
    desc='initial guess for initial_distance')

CruisePhase._add_initial_guess_meta_data(
    InitialGuessState('initial_time'),
    desc='initial guess for initial_time')

CruisePhase._add_initial_guess_meta_data(
    InitialGuessState('altitude'),
    desc='initial guess for altitude')

CruisePhase._add_initial_guess_meta_data(
    InitialGuessState('mach'),
    desc='initial guess for mach')



import unittest

import numpy as np
import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal

from aviary.subsystems.aerodynamics.aerodynamics_builder import CoreAerodynamicsBuilder
from aviary.utils.aviary_values import AviaryValues, get_items
from aviary.models.N3CC.N3CC_data import (
    N3CC, takeoff_subsystem_options, takeoff_subsystem_options_spoilers)
from aviary.variable_info.variables import Aircraft, Dynamic, Mission
from aviary.variable_info.enums import LegacyCode


class TestTakeoffAeroGroup(unittest.TestCase):

    def test_takeoff_aero_group(self):
        prob: om.Problem = make_problem(takeoff_subsystem_options)

        prob.run_model()

        tol = 1e-12

        for (key, (desired, units)) in get_items(_regression_data):
            try:
                actual = prob.get_val(key, units)

                assert_near_equal(actual, desired, tol)

            except ValueError as error:
                msg = f'"{key}": {error!s}'

                raise ValueError(msg) from None

        partials = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partials, atol=1e-10, rtol=1e-12)

    def test_takeoff_aero_group_spoiler(self):
        prob: om.Problem = make_problem(takeoff_subsystem_options_spoilers)

        prob.run_model()

        tol = 1e-12

        for (key, (desired, units)) in get_items(_regression_data_spoiler):
            try:
                actual = prob.get_val(key, units)

                assert_near_equal(actual, desired, tol)

            except ValueError as error:
                msg = f'"{key}": {error!s}'

                raise ValueError(msg) from None

        partials = prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partials, atol=1e-10, rtol=1e-12)


def make_problem(subsystem_options={}):
    '''
    Return a problem that covers all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).
    '''
    nn = 3

    aviary_keys = (
        Mission.Takeoff.DRAG_COEFFICIENT_MIN,
        Aircraft.Wing.ASPECT_RATIO, Aircraft.Wing.HEIGHT,
        Aircraft.Wing.SPAN, Aircraft.Wing.AREA)

    aviary_inputs = AviaryValues(N3CC['inputs'])

    dynamic_inputs = AviaryValues({
        'angle_of_attack': (np.array([0., 2., 6.]), 'deg'),
        Dynamic.Mission.ALTITUDE: (np.array([0., 32., 55.]), 'm'),
        Dynamic.Mission.FLIGHT_PATH_ANGLE: (np.array([0., 0.5, 1.]), 'deg')})

    prob = om.Problem()

    prob.model.add_subsystem(
        "USatm",
        USatm1976Comp(num_nodes=nn),
        promotes_inputs=[("h", Dynamic.Mission.ALTITUDE)],
        promotes_outputs=[
            "rho", ("sos", Dynamic.Mission.SPEED_OF_SOUND), ("temp",
                                                             Dynamic.Mission.TEMPERATURE),
            ("pres", Dynamic.Mission.STATIC_PRESSURE), "viscosity"])

    aero_builder = CoreAerodynamicsBuilder(code_origin=LegacyCode.FLOPS)

    prob.model.add_subsystem(
        name='core_aerodynamics',
        subsys=aero_builder.build_mission(num_nodes=nn,
                                          aviary_inputs=aviary_inputs,
                                          **subsystem_options['core_aerodynamics']),
        promotes_inputs=aero_builder.mission_inputs(
            **subsystem_options['core_aerodynamics']),
        promotes_outputs=aero_builder.mission_outputs(**subsystem_options['core_aerodynamics']))

    prob.model.set_input_defaults(Dynamic.Mission.ALTITUDE, np.zeros(nn), 'm')

    prob.setup(force_alloc_complex=True)

    for key in aviary_keys:
        try:
            val, units = aviary_inputs.get_item(key, ())

        except Exception as error:
            msg = f'"{key}": {error!s}'

            raise ValueError(msg) from None

        prob.set_val(key, val, units)

    for (key, (val, units)) in dynamic_inputs:
        prob.set_val(key, val, units)

    return prob


_units_lift = 'N'
_units_drag = 'N'


# NOTE:
# - results from `generate_regression_data()`
#    - last generated 2023 June 8
# - generate new regression data if, and only if, takeoff aero group is updated with a
#   more trusted implementation
_regression_data = AviaryValues({
    Dynamic.Mission.LIFT: (
        [3028.138891962988, 4072.059743068957, 6240.85493286], _units_lift),
    Dynamic.Mission.DRAG: (
        [434.6285684000267, 481.5245555324278, 586.0976806512001], _units_drag)})

# NOTE:
# - results from `generate_regression_data_spoiler()`
#    - last generated 2023 June 8
# - generate new regression data if, and only if, takeoff aero group is updated with a
#   more trusted implementation
_regression_data_spoiler = AviaryValues({
    Dynamic.Mission.LIFT: (
        [-1367.5937129210124, -323.67286181504335, 1845.1223279759993], _units_lift),
    Dynamic.Mission.DRAG: (
        [895.9091503940268, 942.8051375264279, 1047.3782626452], _units_drag)})


def generate_regression_data():
    '''
    Generate regression data that covers all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).

    Note, spoiler disabled.

    Notes
    -----
    Use this function to generate new regression data if, and only if, ground effect is
    updated with a more trusted implementation.
    '''
    _generate_regression_data(takeoff_subsystem_options)


def generate_regression_data_spoiler():
    '''
    Generate regression data that covers all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).

    Note, spoiler enabled.

    Notes
    -----
    Use this function to generate new regression data if, and only if, ground effect is
    updated with a more trusted implementation.
    '''
    _generate_regression_data(takeoff_subsystem_options_spoilers)


def _generate_regression_data(subsystem_options={}):
    '''
    Generate regression data that covers all three computational ranges:
    - in ground effect (on the ground);
    - in transition (in the air, but near enough to the ground for limited effect);
    - in free flight (no ground effect).

    Note, results are aero builder specific.

    Notes
    -----
    Use this function to generate new regression data if, and only if, ground effect is
    updated with a more trusted implementation.
    '''
    prob: om.Problem = make_problem(subsystem_options)

    prob.run_model()

    lift = prob.get_val(Dynamic.Mission.LIFT, _units_lift)
    drag = prob.get_val(Dynamic.Mission.DRAG, _units_drag)

    prob.check_partials(compact_print=True, method="cs")

    print('lift', *lift, sep=', ')
    print('drag', *drag, sep=', ')


if __name__ == "__main__":
    unittest.main()
    # test = TestTakeoffAeroGroup()
    # test.test_takeoff_aero_group()



import unittest

import aviary.api as av
from aviary.utils.merge_hierarchies import merge_hierarchies

AviaryAircraft = av.Aircraft
AviaryMission = av.Mission


class Aircraft1(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    CG = "aircraft:center_of_gravity"

    class LandingGear(AviaryAircraft.LandingGear):
        SHAPE = 'aircraft:landing_gear:shape'

    class Wing(AviaryAircraft.Wing):
        CHARACTER = 'aircraft:wing:character'

    class Jury:
        MASS = 'aircraft:jury:mass'

    class Engine(AviaryAircraft.Engine):
        COLOR = 'aircraft:engine:color'

    class Fuselage(AviaryAircraft.Fuselage):

        class Nose:
            ECCENTRICITY = 'aircraft:fuselage:nose:eccentricity'

    class Test:
        var1 = '1'

        class SubTest:
            var2 = '2'

            class SubSubTest:
                var3 = '3'

                class SubSubSubTest:
                    var4 = '4'


class Aircraft2(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    MASS = "aircraft:mass"
    CG = 'aircraft:center_of_gravity'

    class HorizontalTail(AviaryAircraft.HorizontalTail):
        # add variable that is in the core hierarchy and give it the right value
        AREA = 'aircraft:horizontal_tail:area'
        MEAN_AERO_CHORD = "aircraft:horizontal_tail:mean_aerodynamic_chord"
        TEST = 'test'

        class Elevator:
            AREA = "aircraft:horizontal_tail:elevator_area_dist"

    class Wing(AviaryAircraft.Wing):

        AERO_CENTER = "aircraft:wing:aerodynamic_center"
        CHORD = "aircraft:wing:chord"
        AIRFOIL_TECHNOLOGY = 'aircraft:wing:airfoil_technology'

        class Flap:
            AREA = "aircraft:wing:flap_area_dist"
            ROOT_CHORD = "aircraft:wing:flap_root_chord_dist"
            SPAN = "aircraft:wing:flap_span_dist"

    class Engine(AviaryAircraft.Engine):
        COLOR = 'aircraft:engine:color'


class Aircraft3(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    class HorizontalTail(AviaryAircraft.HorizontalTail):

        class Elevator:
            SPAN = "aircraft:horizontal_tail:elevator_span"

    class Wing(AviaryAircraft.Wing):

        class Flap:
            CHORD = "aircraft:wing:flap_chord"


class Aircraft4(AviaryAircraft):
    ''' Simple variable hierarchy extension for testing. '''

    class HorizontalTail(AviaryAircraft.HorizontalTail):

        class Elevator:
            SPAN = "wrong_value"


class SimpleHierarchy:
    class InnerClass:
        var1 = 'dummy'


class SimpleHierarchy2:
    class SubOne:
        var_sub_one = 'dummy1'

    class SubTwo:
        var_sub_two = 'dummy2'

    dummy_var = 'dummy3'


class Mission1(AviaryMission):
    dummy = 'dummy'


class DeepSub:
    class Test:
        class SubTest:
            class Sub2Test:
                class Sub3Test:
                    class Sub4Test:
                        class Sub5Test:
                            class Sub6Test:
                                class Sub7Test:
                                    class Sub8Test:
                                        class Sub9Test:
                                            class Sub10Test:
                                                var10 = 'dummy10'


merge_combo1 = [SimpleHierarchy, Aircraft1, Aircraft2]
merge_combo2 = [Aircraft1, Aircraft2, SimpleHierarchy]
merge_combo3 = [Aircraft1, Aircraft2, SimpleHierarchy, Mission1]
merge_combo4 = [SimpleHierarchy2, SimpleHierarchy]
merge_combo5 = [SimpleHierarchy, DeepSub]
merge_combo6 = [DeepSub, SimpleHierarchy]
merge_combo7 = [Aircraft2, Aircraft3]
merge_combo8 = [Aircraft3, Aircraft4]


class MergeHierarchiesTest(unittest.TestCase):
    def test_merge1(self):
        self.maxDiff = None

        merged = merge_hierarchies(merge_combo1)

        # check for presence of all variables from input hierarchies
        self.assertEqual(merged.InnerClass.var1, 'dummy')
        self.assertEqual(merged.CG, 'aircraft:center_of_gravity')
        self.assertEqual(merged.LandingGear.SHAPE, 'aircraft:landing_gear:shape')
        self.assertEqual(merged.Wing.CHARACTER, 'aircraft:wing:character')
        self.assertEqual(merged.Jury.MASS, 'aircraft:jury:mass')
        self.assertEqual(merged.Engine.COLOR, 'aircraft:engine:color')
        self.assertEqual(merged.Fuselage.Nose.ECCENTRICITY,
                         'aircraft:fuselage:nose:eccentricity')
        self.assertEqual(merged.Test.var1, '1')
        self.assertEqual(merged.Test.SubTest.var2, '2')
        self.assertEqual(merged.Test.SubTest.SubSubTest.var3, '3')
        self.assertEqual(merged.Test.SubTest.SubSubTest.SubSubSubTest.var4, '4')
        self.assertEqual(merged.MASS, 'aircraft:mass')
        self.assertEqual(merged.HorizontalTail.AREA, 'aircraft:horizontal_tail:area')
        self.assertEqual(merged.HorizontalTail.MEAN_AERO_CHORD,
                         'aircraft:horizontal_tail:mean_aerodynamic_chord')
        self.assertEqual(merged.HorizontalTail.TEST, 'test')
        self.assertEqual(merged.HorizontalTail.Elevator.AREA,
                         'aircraft:horizontal_tail:elevator_area_dist')
        self.assertEqual(merged.Wing.AERO_CENTER, 'aircraft:wing:aerodynamic_center')
        self.assertEqual(merged.Wing.CHORD, 'aircraft:wing:chord')
        self.assertEqual(merged.Wing.AIRFOIL_TECHNOLOGY,
                         'aircraft:wing:airfoil_technology')
        self.assertEqual(merged.Wing.Flap.AREA, 'aircraft:wing:flap_area_dist')
        self.assertEqual(merged.Wing.Flap.ROOT_CHORD,
                         'aircraft:wing:flap_root_chord_dist')
        self.assertEqual(merged.Wing.Flap.SPAN, 'aircraft:wing:flap_span_dist')
        self.assertEqual(merged.Engine.COLOR, 'aircraft:engine:color')
        self.assertEqual(merged.MASS, 'aircraft:mass')

        # check for presence of variables from original Aviary core hierarchy
        self.assertEqual(merged.AirConditioning.MASS, 'aircraft:air_conditioning:mass')
        self.assertEqual(merged.AntiIcing.MASS_SCALER, 'aircraft:anti_icing:mass_scaler')
        self.assertEqual(merged.APU.MASS, 'aircraft:apu:mass')
        self.assertEqual(merged.Canard.AREA, 'aircraft:canard:area')
        self.assertEqual(merged.CrewPayload.MISC_CARGO,
                         'aircraft:crew_and_payload:misc_cargo')
        self.assertEqual(merged.Design.EMPTY_MASS, 'aircraft:design:empty_mass')
        self.assertEqual(merged.Wing.BWB_AFTBODY_MASS, 'aircraft:wing:bwb_aft_body_mass')
        self.assertEqual(merged.Hydraulics.MASS, 'aircraft:hydraulics:mass')

    def test_merge2(self):
        self.maxDiff = None

        merged = merge_hierarchies(merge_combo2)

        # check for presence of all variables from input hierarchies
        self.assertEqual(merged.InnerClass.var1, 'dummy')
        self.assertEqual(merged.CG, 'aircraft:center_of_gravity')
        self.assertEqual(merged.LandingGear.SHAPE, 'aircraft:landing_gear:shape')
        self.assertEqual(merged.Wing.CHARACTER, 'aircraft:wing:character')
        self.assertEqual(merged.Jury.MASS, 'aircraft:jury:mass')
        self.assertEqual(merged.Engine.COLOR, 'aircraft:engine:color')
        self.assertEqual(merged.Fuselage.Nose.ECCENTRICITY,
                         'aircraft:fuselage:nose:eccentricity')
        self.assertEqual(merged.Test.var1, '1')
        self.assertEqual(merged.Test.SubTest.var2, '2')
        self.assertEqual(merged.Test.SubTest.SubSubTest.var3, '3')
        self.assertEqual(merged.Test.SubTest.SubSubTest.SubSubSubTest.var4, '4')
        self.assertEqual(merged.MASS, 'aircraft:mass')
        self.assertEqual(merged.HorizontalTail.AREA, 'aircraft:horizontal_tail:area')
        self.assertEqual(merged.HorizontalTail.MEAN_AERO_CHORD,
                         'aircraft:horizontal_tail:mean_aerodynamic_chord')
        self.assertEqual(merged.HorizontalTail.TEST, 'test')
        self.assertEqual(merged.HorizontalTail.Elevator.AREA,
                         'aircraft:horizontal_tail:elevator_area_dist')
        self.assertEqual(merged.Wing.AERO_CENTER, 'aircraft:wing:aerodynamic_center')
        self.assertEqual(merged.Wing.CHORD, 'aircraft:wing:chord')
        self.assertEqual(merged.Wing.AIRFOIL_TECHNOLOGY,
                         'aircraft:wing:airfoil_technology')
        self.assertEqual(merged.Wing.Flap.AREA, 'aircraft:wing:flap_area_dist')
        self.assertEqual(merged.Wing.Flap.ROOT_CHORD,
                         'aircraft:wing:flap_root_chord_dist')
        self.assertEqual(merged.Wing.Flap.SPAN, 'aircraft:wing:flap_span_dist')
        self.assertEqual(merged.Engine.COLOR, 'aircraft:engine:color')
        self.assertEqual(merged.MASS, 'aircraft:mass')

        # check for presence of variables from original Aviary core hierarchy
        self.assertEqual(merged.AirConditioning.MASS, 'aircraft:air_conditioning:mass')
        self.assertEqual(merged.AntiIcing.MASS_SCALER, 'aircraft:anti_icing:mass_scaler')
        self.assertEqual(merged.APU.MASS, 'aircraft:apu:mass')
        self.assertEqual(merged.Canard.AREA, 'aircraft:canard:area')
        self.assertEqual(merged.CrewPayload.MISC_CARGO,
                         'aircraft:crew_and_payload:misc_cargo')
        self.assertEqual(merged.Design.EMPTY_MASS, 'aircraft:design:empty_mass')
        self.assertEqual(merged.Wing.BWB_AFTBODY_MASS, 'aircraft:wing:bwb_aft_body_mass')
        self.assertEqual(merged.Hydraulics.MASS, 'aircraft:hydraulics:mass')

    def test_merge3(self):
        self.maxDiff = None
        combo3_error_msg = "You have attempted to merge together variable hierarchies that subclass from different superclasses. 'Aircraft1' is a subclass of '<class 'aviary.variable_info.variables.Aircraft'>' and 'Mission1' is a subclass of '<class 'aviary.variable_info.variables.Mission'>'."

        with self.assertRaises(ValueError) as cm:
            merged = merge_hierarchies(merge_combo3)
        self.assertEqual(str(cm.exception), combo3_error_msg)

    def test_merge4(self):
        self.maxDiff = None

        merged = merge_hierarchies(merge_combo4)

        self.assertEqual(merged.SubOne.var_sub_one, 'dummy1')
        self.assertEqual(merged.SubTwo.var_sub_two, 'dummy2')
        self.assertEqual(merged.dummy_var, 'dummy3')
        self.assertEqual(merged.InnerClass.var1, 'dummy')

    def test_merge5(self):
        self.maxDiff = None

        merged = merge_hierarchies(merge_combo5)

        self.assertEqual(merged.InnerClass.var1, 'dummy')
        self.assertEqual(
            merged.Test.SubTest.Sub2Test.Sub3Test.Sub4Test.Sub5Test.Sub6Test.Sub7Test.Sub8Test.Sub9Test.Sub10Test.var10, 'dummy10')

    def test_merge6(self):
        self.maxDiff = None

        merged = merge_hierarchies(merge_combo6)

        self.assertEqual(merged.InnerClass.var1, 'dummy')
        self.assertEqual(
            merged.Test.SubTest.Sub2Test.Sub3Test.Sub4Test.Sub5Test.Sub6Test.Sub7Test.Sub8Test.Sub9Test.Sub10Test.var10, 'dummy10')

    def test_merge7(self):
        self.maxDiff = None

        merged = merge_hierarchies(merge_combo7)

        self.assertEqual(merged.MASS, 'aircraft:mass')
        self.assertEqual(merged.CG, 'aircraft:center_of_gravity')
        self.assertEqual(merged.HorizontalTail.AREA, 'aircraft:horizontal_tail:area')
        self.assertEqual(merged.HorizontalTail.MEAN_AERO_CHORD,
                         'aircraft:horizontal_tail:mean_aerodynamic_chord')
        self.assertEqual(merged.HorizontalTail.TEST, 'test')
        self.assertEqual(merged.HorizontalTail.Elevator.AREA,
                         'aircraft:horizontal_tail:elevator_area_dist')
        self.assertEqual(merged.HorizontalTail.Elevator.SPAN,
                         'aircraft:horizontal_tail:elevator_span')
        self.assertEqual(merged.Wing.AERO_CENTER, 'aircraft:wing:aerodynamic_center')
        self.assertEqual(merged.Wing.CHORD, 'aircraft:wing:chord')
        self.assertEqual(merged.Wing.AIRFOIL_TECHNOLOGY,
                         'aircraft:wing:airfoil_technology')
        self.assertEqual(merged.Wing.Flap.CHORD, 'aircraft:wing:flap_chord')
        self.assertEqual(merged.Wing.Flap.ROOT_CHORD,
                         'aircraft:wing:flap_root_chord_dist')
        self.assertEqual(merged.Wing.Flap.AREA, 'aircraft:wing:flap_area_dist')
        self.assertEqual(merged.Wing.Flap.SPAN, 'aircraft:wing:flap_span_dist')
        self.assertEqual(merged.Engine.COLOR, 'aircraft:engine:color')

    def test_merge8(self):
        self.maxDiff = None
        combo8_error_msg = "You have attempted to merge two variable hierarchies together that have the same variable with a different string name associated to it. The offending variable is 'SPAN'. In 'Aircraft3.HorizontalTail.Elevator' it has a value of 'aircraft:horizontal_tail:elevator_span' and in 'Aircraft4.HorizontalTail.Elevator' it has a value of 'wrong_value'."

        with self.assertRaises(ValueError) as cm:
            merged = merge_hierarchies(merge_combo8)
        self.assertEqual(str(cm.exception), combo8_error_msg)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import openmdao.api as om

from aviary.variable_info.variables import Dynamic, Mission


class MachNumber(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input(Dynamic.Mission.VELOCITY, val=np.ones(nn),
                       desc='true airspeed', units='m/s')
        self.add_input(Dynamic.Mission.SPEED_OF_SOUND, val=np.ones(
            nn), desc='speed of sound', units='m/s')
        self.add_output(Dynamic.Mission.MACH, val=np.ones(
            nn), desc='current Mach number', units='unitless')

    def compute(self, inputs, outputs):
        sos = inputs[Dynamic.Mission.SPEED_OF_SOUND]
        velocity = inputs[Dynamic.Mission.VELOCITY]

        outputs[Dynamic.Mission.MACH] = velocity/sos

    def setup_partials(self):
        arange = np.arange(self.options['num_nodes'])
        self.declare_partials(
            Dynamic.Mission.MACH, [Dynamic.Mission.SPEED_OF_SOUND, Dynamic.Mission.VELOCITY], rows=arange, cols=arange)

    def compute_partials(self, inputs, J):
        sos = inputs[Dynamic.Mission.SPEED_OF_SOUND]
        velocity = inputs[Dynamic.Mission.VELOCITY]

        J[Dynamic.Mission.MACH, Dynamic.Mission.VELOCITY] = 1/sos
        J[Dynamic.Mission.MACH, Dynamic.Mission.SPEED_OF_SOUND] = -velocity/sos**2



import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.mass.flops_based.distributed_prop import \
    distributed_engine_count_factor
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class TransportFuelSystemMass(om.ExplicitComponent):
    """
    Component to calculate fuel system mass of Transports. The methodology
    is based on the FLOPS weight equations, modified to output mass instead
    of weight.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1.0)

        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, 0.0)

        add_aviary_output(self, Aircraft.Fuel.FUEL_SYSTEM_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        scaler = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]
        capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        num_eng = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_eng_fact = distributed_engine_count_factor(num_eng)
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        outputs[Aircraft.Fuel.FUEL_SYSTEM_MASS] = (
            1.07 * capacity**0.58
            * num_eng_fact**0.43 * max_mach**0.34 * scaler) / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        scaler = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]
        capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        num_eng = aviary_options.get_val(Aircraft.Propulsion.TOTAL_NUM_ENGINES)
        num_eng_fact = distributed_engine_count_factor(num_eng)
        max_mach = aviary_options.get_val(Mission.Constraints.MAX_MACH)

        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER] = (
            1.07 * capacity**0.58 * num_eng_fact**0.43 * max_mach**0.34 / GRAV_ENGLISH_LBM)

        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.TOTAL_CAPACITY] = (
            1.07 * 0.58 * capacity**-0.42 * num_eng**0.43 * max_mach**0.34
            * scaler / GRAV_ENGLISH_LBM)


class AltFuelSystemMass(om.ExplicitComponent):
    """
    Component for an alternate way to calculate fuel system mass. The methodology
    is based on the FLOPS weight equations, modified to output mass instead
    of weight.
    """

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Fuel.TOTAL_CAPACITY, val=0.0)

        add_aviary_input(self, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER, val=1.0)

        add_aviary_output(self, Aircraft.Fuel.FUEL_SYSTEM_MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        number_of_fuel_tanks = aviary_options.get_val(Aircraft.Fuel.NUM_TANKS)
        total_fuel_capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        scaler = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]

        fuel_sys_weight = (
            978.6 *
            (number_of_fuel_tanks / 13.0)
            + 2283.4
            * (total_fuel_capacity / 208100.0)**(2.0 / 3.0)
            + 350.0 + 0.00029 * total_fuel_capacity) * scaler

        outputs[Aircraft.Fuel.FUEL_SYSTEM_MASS] = \
            fuel_sys_weight / GRAV_ENGLISH_LBM

    def compute_partials(self, inputs, J):
        aviary_options: AviaryValues = self.options['aviary_options']
        number_of_fuel_tanks = aviary_options.get_val(Aircraft.Fuel.NUM_TANKS)
        total_fuel_capacity = inputs[Aircraft.Fuel.TOTAL_CAPACITY]
        scaler = inputs[Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER]

        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.TOTAL_CAPACITY] = (
            (2283.4 * (2.0 / 3.0) * (1 / 208100.0)**(2.0 / 3.0)
             * total_fuel_capacity ** (-1.0/3.0)
             + 0.00029) * scaler / GRAV_ENGLISH_LBM
        )

        J[Aircraft.Fuel.FUEL_SYSTEM_MASS, Aircraft.Fuel.FUEL_SYSTEM_MASS_SCALER] = (
            0.3 * 3262.0 *
            (number_of_fuel_tanks / 13.0)
            + 0.7 * 3262.0
            * (total_fuel_capacity / 208100.0)**(2.0 / 3.0)
            + 350.0 + 0.00029 * total_fuel_capacity) / GRAV_ENGLISH_LBM



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class PaintMass(om.ExplicitComponent):
    '''
    Calculates the mass of paint based on total wetted area.
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Design.TOTAL_WETTED_AREA, val=0.0)

        add_aviary_input(self, Aircraft.Paint.MASS_PER_UNIT_AREA, val=0.0)

        add_aviary_output(self, Aircraft.Paint.MASS, val=0.0)

    def setup_partials(self):
        self.declare_partials('*', '*')

    def compute(
        self, inputs, outputs, discrete_inputs=None, discrete_outputs=None
    ):
        wetted_area = inputs[Aircraft.Design.TOTAL_WETTED_AREA]
        mass_per_area = inputs[Aircraft.Paint.MASS_PER_UNIT_AREA]

        outputs[Aircraft.Paint.MASS] = wetted_area * mass_per_area

    def compute_partials(self, inputs, J):
        wetted_area = inputs[Aircraft.Design.TOTAL_WETTED_AREA]
        mass_per_area = inputs[Aircraft.Paint.MASS_PER_UNIT_AREA]

        J[Aircraft.Paint.MASS, Aircraft.Design.TOTAL_WETTED_AREA] = \
            mass_per_area

        J[Aircraft.Paint.MASS, Aircraft.Paint.MASS_PER_UNIT_AREA] = \
            wetted_area



import numpy as np
import openmdao.api as om
from openmdao.utils import units
from scipy import interpolate
from simupy.block_diagram import DEFAULT_INTEGRATOR_OPTIONS, SimulationMixin
from simupy.systems import DynamicalSystem

from aviary.mission.gasp_based.ode.params import ParamPort


def add_SGM_required_inputs(group: om.Group, inputs_to_add: dict):
    blank_execcomp = om.ExecComp()
    for input, details in inputs_to_add.items():
        blank_execcomp.add_input(input, **details)
    group.add_subsystem('SGM_required_inputs',
                        blank_execcomp,
                        promotes_inputs=list(inputs_to_add.keys()),
                        )


# Subproblem used as a basis for forward in time integration phases.
class SimuPyProblem(SimulationMixin):
    def __init__(
        self,
        ode,
        time_independent=False,
        t_name="t_curr",
        states=None,
        alternate_state_names=None,
        blocked_state_names=None,
        alternate_state_rate_names=None,
        parameters=None,
        outputs=None,
        controls=None,
        include_state_outputs=False,
        rate_suffix="_rate",
        DEBUG=False,
        max_allowable_time=1_000_000,
        adjoint_int_opts=DEFAULT_INTEGRATOR_OPTIONS.copy(),

    ):
        """
        states: a dictionary of the form {state_name:{'units':unit, 'rate':state_rate_name, 'rate_units':state_rate_units}}
        alternate_state_names: a dictionary of the form {state_to_replace:new_state_name}
        blocked_state_names: a list of the form [state_name_to_block]
        alternate_state_rate_names: a dictionary of the form {state_to_modify:new_state_rate_name}
        parameters: a dictionary of the form {parameter_name:unit}
        outputs: a dictionary of the form {output_name:unit}
        controls: a dictionary of the form {control_name:unit}
        include_state_outputs : automatically add the state to the input
        works well for auto-parsed naming, does not check for duplication before adding
        states, parameters, outputs, and controls can also be input as a list of keys for the dictionary 
        """

        default_om_list_args = dict(prom_name=True, val=False,
                                    out_stream=None, units=True)

        self.DEBUG = DEBUG
        self.max_allowable_time = max_allowable_time
        self.adjoint_int_opts = adjoint_int_opts
        self.adjoint_int_opts['nsteps'] = 5000
        self.adjoint_int_opts['name'] = "dop853"

        self.dt = 0.0
        prob = om.Problem()
        prob.model.add_subsystem(
            "ODE_group",
            ode,
            promotes=["*"],
        )
        self.ode = ode

        self.prob = prob
        prob.setup(check=False, force_alloc_complex=True)
        prob.final_setup()

        self.time_independent = time_independent
        if type(states) is list:
            states = {state: {'units': None, 'rate': state+rate_suffix, 'rate_units': None}
                      for state in states}
        if type(parameters) is list:
            parameters = {parameter: None for parameter in parameters}
        if type(outputs) is list:
            outputs = {output: None for output in outputs}
        if controls is None:
            controls = {}
        elif type(controls) is list:
            controls = {control: None for control in controls}
        if alternate_state_names is None:
            alternate_state_names = {}
        if alternate_state_rate_names is None:
            alternate_state_rate_names = {}

        data = prob.model.list_inputs(includes=controls.keys(), **default_om_list_args)
        control_data = {val.pop('prom_name'): val for key, val in data}
        for control_name, control_units in controls.items():
            if control_units is None:
                controls[control_name] = control_data[control_name]["units"]
        self.controls = controls

        if (
            states is None
            or parameters is None
            or outputs is None  # or
            # event_names is None
        ):
            data = prob.model.list_outputs(**default_om_list_args)
            model_outputs = [val['prom_name'] for key, val in data]
            data = prob.model.list_inputs(**default_om_list_args)
            model_inputs = [val['prom_name'] for key, val in data]

        if states is None:
            states = {
                output.replace(rate_suffix, ""): {'units': None, 'rate': output}
                for output in model_outputs
                if output.endswith(rate_suffix)
                or output in alternate_state_names.keys()
            }

            if blocked_state_names is not None:
                for name in blocked_state_names:
                    if name in states.keys():
                        states.pop(name)
            if alternate_state_names:
                for state_key, val in alternate_state_names.items():
                    if state_key in states.keys():  # Used to rename an existing state
                        states[val] = states.pop(state_key)
                    else:  # Used to add a new state
                        states[val]: {'units': None, 'rate': val+rate_suffix}

        if alternate_state_rate_names:
            for state_name, val in alternate_state_rate_names.items():
                if state_name in states.keys():
                    states[state_name]['rate'] = val

        state_rate_names = [val['rate'] for _, val in states.items()]
        data = prob.model.list_outputs(includes=state_rate_names, **default_om_list_args)
        rate_data = {val.pop('prom_name'): val for key, val in data}
        for state_name, state_data in states.items():
            if state_data['rate_units'] is None:
                state_data['rate_units'] = rate_data[state_data['rate']]["units"]
            if state_data['units'] is None:
                rate_unit = rate_data[state_data['rate']]["units"]
                states[state_name]['units'] = units.simplify_unit("s*" + rate_unit)

        if parameters is None:
            parameters = {
                inp: None
                for inp in set(model_inputs)
                if inp not in list(states.keys()) + list(controls.keys()) + [t_name]
            }

        data = prob.model.list_inputs(includes=parameters.keys(), **default_om_list_args)
        parameter_data = {val.pop('prom_name'): val for key, val in data}
        for parameter_name, parameter_units in parameters.items():
            if parameter_units is None:
                parameters[parameter_name] = parameter_data[parameter_name]["units"]

        if outputs is None:
            outputs = {
                outp: None
                for outp in outputs
                if outp not in state_rate_names  # + event_names
            }

        for output_name, output_units in outputs.items():
            if output_units is None:
                outputs[output_name] = next(iter(prob.model.get_io_metadata(
                    includes=[output_name],
                    metadata_keys=["units"]
                ).values()))["units"]

        if include_state_outputs or outputs == {}:  # prevent empty outputs
            outputs = states + outputs

        self.t_name = t_name
        self.states = states

        self.parameters = parameters
        self.outputs = outputs

        self.dim_state = len(states)
        self.dim_output = len(outputs)
        self.dim_input = len(controls)
        self.dim_parameters = len(parameters)
        # TODO: add defensive checks to make sure dimensions match in both setup and
        # calls

        if DEBUG or True:
            om.n2(prob, outfile="n2_simupy_problem.html", show_browser=False)
            with open('input_list_simupy.txt', 'w') as outfile:
                prob.model.list_inputs(out_stream=outfile,)
            print(states)

    @property
    def time(self):
        return self.prob.get_val(self.t_name)[0]

    @time.setter
    def time(self, value):
        if self.time_independent or self.time == value:
            return
        self.prob.set_val(self.t_name, value)

    @property
    def state(self):
        return np.array(
            [
                self.prob.get_val(state_name, units=state_data['units'])[0]
                for state_name, state_data in self.states.items()
            ]
        )

    @state.setter
    def state(self, value):
        if np.all(self.state == value):
            return
        for state_name, elem_val in zip(
            self.states.keys(), value
        ):
            self.prob.set_val(state_name, elem_val,
                              units=self.states[state_name]['units'])

    def compute_along_traj(self, ts, xs):
        self.prob.set_val(self.t_name, ts)
        for state_name, elem_val in zip(self.states.keys(), xs.T):
            self.prob.set_val(state_name, elem_val,
                              units=self.states[state_name]['units'])

        self.prob.run_model()

    @property
    def control(self):
        return np.array(
            [
                self.prob.get_val(control_name, units=unit)[0]
                for control_name, unit in self.controls.items()
            ]
        )

    @control.setter
    def control(self, value):
        if value is None:
            value = np.array([])
        if (self.control.size == value.size) and np.all(self.control == value):
            return
        for control_name, elem_val in zip(
            self.controls, value
        ):
            self.prob.set_val(control_name, elem_val, units=self.controls[control_name])

    @property
    def parameter(self):
        return np.array(
            [
                self.prob.get_val(parameter_name, units=unit)[0]
                for parameter_name, unit in self.parameters.items()
            ]
        )

    @parameter.setter
    def parameter(self, value):
        if np.all(self.parameter == value):
            return
        for parameter_name, elem_val in zip(
            self.paramaters, value
        ):
            self.prob.set_val(parameter_name, elem_val)

    @property
    def state_rate(self):
        return np.array(
            [
                self.prob.get_val(state_data['rate'], units=state_data['rate_units'])[0]
                for _, state_data in self.states.items()
            ]
        )

    @property
    def output(self):
        return np.array(
            [
                self.prob.get_val(output_name, units=unit)[0]
                for output_name, unit in self.outputs.items()
            ]
        )

    @property
    def events(self):
        return np.array(
            [
                self.prob.get_val(event_name, units=unit)[0]
                for event_name, unit in zip(self.event_names, self.event_units)
            ]
        )

    @property
    def compute(self):
        # TODO: create cache -- t, x (optionally u) as hash
        # I guess just set_val for every input and output?
        # for compute_totals, force run_model. anywhere else that needs forced
        # run_model?
        return self.prob.run_model

    @property
    def compute_totals(self):
        return self.prob.compute_totals

    def state_equation_function(self, t, x, u=None):
        self.time = t
        self.state = x
        self.control = u
        self.compute()
        return self.state_rate

    def output_equation_function(self, t, x):
        if self.output_nan:
            return np.ones(self.dim_output) * np.nan
        self.time = t
        self.state = x
        self.compute()
        return self.output

    def prepare_to_integrate(self, t0, x0):
        self.output_nan = False
        # self.time = t0
        # self.state = x0
        # self.prob.run_model()
        return self.output_equation_function(t0, x0)

    def update_equation_function(self, t, x, event_channels=None):
        self.output_nan = True
        return x

    @property
    def get_val(self):
        return self.prob.get_val

    @property
    def set_val(self):
        return self.prob.set_val


class SGMTrajBase(om.ExplicitComponent):
    def initialize(self):
        # needs to get passed to each ODE
        # TODO: param_dict
        self.options.declare("param_dict",
                             default=ParamPort.param_data)
        self.DEBUG = False
        self.max_allowable_time = 1_000_000
        self.adjoint_int_opts = DEFAULT_INTEGRATOR_OPTIONS.copy()
        self.adjoint_int_opts['nsteps'] = 5000
        self.adjoint_int_opts['name'] = "dop853"

    def setup_params(
            self,
            ODEs,
            promote_all_auto_ivc=False,
            traj_final_state_output=None,
            traj_promote_final_output=None,
            traj_promote_initial_input=None,
            traj_initial_state_input=None,
            traj_event_trigger_input=None,
    ):
        """
        API requirements:
            pass ODE's,
            next_problem = f(current_problem, current_result)
            initial_state/time/etc 
            next_state from last state/output/event information

            pass in terminal and integrand output functions with derivatives (components)
            -- anything special for final state, final time?
            declare initial state(s) as parameters to take derivative wrt
            assume all other inputs are parameters for deriv? 
        """
        if traj_final_state_output is None:
            traj_final_state_output = []
        if traj_promote_final_output is None:
            traj_promote_final_output = []
        if traj_promote_initial_input is None:
            traj_promote_initial_input = {}
        if traj_initial_state_input is None:
            traj_initial_state_input = []
        if traj_event_trigger_input is None:
            traj_event_trigger_input = []

        if promote_all_auto_ivc:
            for ode in ODEs:
                ode: SimuPyProblem
                ode.auto_ivc_vars = ode.prob.model.list_inputs(
                    is_indep_var=True, units=True, out_stream=None)
                for abs_name, data in ode.auto_ivc_vars:
                    prom_name = data.pop('prom_name')
                    if '.' in prom_name:
                        continue
                    traj_promote_initial_input[prom_name] = data

        self.traj_promote_initial_input = {
            **self.options["param_dict"], **traj_promote_initial_input}
        for name, kwargs in self.traj_promote_initial_input.items():
            self.add_input(name, **kwargs)

        final_suffix = "_final"
        self.traj_final_state_output = {
            final_state_output: {
                **dict(
                    name=final_state_output+final_suffix,
                    state_name=final_state_output,
                ),
                **self.add_output(
                    final_state_output+final_suffix,
                    units=ODEs[-1].states[final_state_output]['units'],
                )
            }
            for final_state_output in traj_final_state_output
        }
        self.traj_promote_final_output = {
            promoted_final_output: {
                **dict(
                    name=promoted_final_output+final_suffix,
                    output_name=promoted_final_output,
                ),
                **self.add_output(
                    promoted_final_output+final_suffix,
                    units=ODEs[-1].outputs[promoted_final_output],
                ),
            }
            for promoted_final_output in traj_promote_final_output
        }
        self.all_traj_outputs = {
            **self.traj_final_state_output,
            **self.traj_promote_final_output,
        }
        initial_suffix = "_initial"
        self.traj_initial_state_input = {
            initial_state_input: {
                **dict(name=initial_state_input+initial_suffix),
                **self.add_input(
                    initial_state_input+initial_suffix,
                    units=ODEs[0].states[initial_state_input]['units'],
                )
            }
            for initial_state_input in traj_initial_state_input
        }

        # TODO: assumes state, not output
        trigger_suffix = "trigger"
        self.traj_event_trigger_input = {
            event_trigger_input: {
                **dict(name="_".join([
                    event_trigger_input[0].__class__.__name__,
                    event_trigger_input[1],
                    trigger_suffix
                ])),
                **self.add_input(
                    "_".join([
                        event_trigger_input[0].__class__.__name__,
                        event_trigger_input[1],
                        trigger_suffix
                    ]),
                    units=event_trigger_input[0].states[event_trigger_input[1]]['units'],
                )
            }
            for event_trigger_input in traj_event_trigger_input
        }
        self.ODEs = ODEs
        self.declare_partials(["*"], ["*"],)

    def compute_params(self, inputs):
        for input in self.traj_promote_initial_input.keys():
            for ode in self.ODEs:
                try:
                    ode.set_val(input, inputs[input])
                except KeyError:
                    if self.DEBUG:
                        print(
                            "*** Input not found:",
                            ode,
                            input
                        )
                    pass

    def compute_traj_loop(self, first_problem, inputs, outputs, t0=0., state0=None):
        if self.DEBUG:
            print("initializing compute_traj_loop")
        sim_results = []
        sim_problems = [first_problem]
        t = t0
        if state0 is not None:
            state = state0
        else:
            state = np.array([
                inputs[state_name+"_initial"].squeeze()
                if state_name in self.traj_initial_state_input
                else 0.
                for state_name in first_problem.states.keys()
            ]).squeeze()

        while True:
            current_problem = sim_problems[-1]
            current_problem.initial_condition = state

            sim_result = current_problem.simulate(
                (t, self.max_allowable_time),
            )
            if sim_result.t.shape[0] == 2:
                print("\n"*3, "IMMEDIATE PHASE TERMINATION", current_problem, "\n"*2)
            sim_results.append(sim_result)

            t = sim_result.t[-1]
            x = sim_result.x[-1, :]

            # TODO: is there a better way to do this? Perhaps don't use for loop -- use
            # while True ?Ij
            try:
                try_next_problem = (yield current_problem, sim_result)
            except GeneratorExit:
                if self.DEBUG:
                    print("stop iteration 1")
                break

            if try_next_problem is not None:
                next_problem = try_next_problem
            else:
                try:
                    next_problem = (yield current_problem, sim_result)
                except GeneratorExit:
                    if self.DEBUG:
                        print("stop iteration 2")
                    break

                if self.DEBUG:
                    print(" was on problem:", current_problem,
                          "\n got back:", next_problem)
            # compute the output at the final condition to make sure all outputs are current
            current_problem.output_equation_function(t, x)
            state = np.array(
                [
                    current_problem.prob.get_val(state_name, units=state_data['units'])
                    for state_name, state_data in next_problem.states.items()
                ]
            ).squeeze()
            sim_problems.append(next_problem)

        if self.DEBUG:
            print("ended loop")

        # wrap main loop
        self.sim_results = sim_results
        self.sim_problems = sim_problems

        # trajectory-specific outputs
        for output in self.traj_final_state_output:
            output_name = self.traj_final_state_output[output]["name"]
            state_name = self.traj_final_state_output[output]["state_name"]

            outputs[output_name] = sim_results[-1].x[
                -1,
                list(sim_problems[-1].states.keys()).index(state_name)
            ]

        for output in self.traj_promote_final_output:
            promoted_name = self.traj_promote_final_output[output]["name"]
            output_name = self.traj_promote_final_output[output]["output_name"]

            outputs[promoted_name] = sim_results[-1].y[
                -1,
                list(sim_problems[-1].outputs.keys()).index(output_name)
            ]

        self.last_inputs = np.array(list(inputs.values()))

    def compute_partials(self, inputs, J):
        self.compute_params(inputs)
        # defensive check -- should really make sure ALL inputs are the same, need a
        # deep copy
        # just calling compute_params doesn't fix it -- possibly different trajectory!

        if np.any(self.last_inputs != np.array(list(inputs.values()))):
            raise ValueError(
                "Attempting to run compute_partials when the last compute"
                " inputs did not match",
            )

        param_dict = self.options["param_dict"]

        # assume the first problem has the most states?
        costate_reses = {output: [] for output in self.all_traj_outputs}
        tf_total = self.sim_results[-1].t[-1]

        next_res = self.sim_results[-1]
        next_prob = self.sim_problems[-1]

        df_dxs = []
        df_dparams = []
        dg_dxs = []
        f_minuses = []
        f_pluses = [np.zeros(self.sim_problems[-1].dim_state)]
        state_updates = [next_res.x[-1, :]]
        dh_dxs = [np.eye(next_prob.dim_state)]

        dh_dparams = [np.zeros((next_prob.dim_state, len(param_dict)))]

        # keep directionality of forward for plant, backward for adjoint by caching
        # everything in order to ensure fewest number of ODE calls with smallest
        # step-size sum(abs(x[i]-x[i-1]), )
        costate_ics = []
        param_derivs = []

        for output in self.all_traj_outputs:
            output_name = self.all_traj_outputs[output]["name"]
            costate = np.zeros(next_prob.dim_state)
            param_deriv = np.zeros(len(param_dict))

            if output in self.traj_final_state_output:
                costate[list(next_prob.states.keys()).index(output)] = 1.
            else:  # in self.traj_promote_final_output

                next_prob.state_equation_function(next_res.t[-1], next_res.x[-1, :])
                costate[:] = next_prob.compute_totals(
                    output,
                    next_prob.states.keys(),
                    return_format='array'
                ).squeeze()

                param_deriv[:] = next_prob.compute_totals(
                    output,
                    list(param_dict.keys()),
                    return_format='array'
                ).squeeze()

            param_derivs.append(param_deriv)
            costate_ics.append(costate)

        # pre-compute data for adjoint
        for phase_idx, res, prob in zip(
            range(len(self.sim_results), 0, -1),
            self.sim_results[::-1],
            self.sim_problems[::-1],
        ):
            # build time-varying co-state matrix
            df_dx_data = np.empty(res.x.shape + (res.x.shape[-1],))

            if param_dict:
                df_dparam_data = np.empty(res.x.shape + (len(param_dict),))

            last_res_idx = res.t.shape[0] - 1

            num_active_event_channels = 0

            state_rate = prob.state_equation_function(res.t[-1], res.x[-1, :])
            if (prob is not self.sim_problems[0]):
                f_minuses.append(state_rate)

            for channel_idx, channel_name in enumerate(prob.event_channel_names):

                if np.argmin(np.abs(res.e[-1, :])) not in [channel_idx]:
                    continue

                num_active_event_channels += 1
                dg_dx = np.zeros((1, prob.dim_state))

                if channel_name in prob.states.keys():
                    dg_dx[0, list(prob.states.keys()).index(channel_name)] = 1.
                else:
                    dg_dx[0, :] = prob.compute_totals(
                        [channel_name],
                        prob.states.keys(),
                        return_format='array'
                    )

                # TODO: actually save dg_dts
                if channel_name == prob.t_name:
                    dg_dt = 1.
                else:
                    dg_dt = 0.

            dg_dxs.append(dg_dx)

            if num_active_event_channels != 1:
                raise ValueError("Somehow ended a phase without an event? assume "
                                 "time in the future?? but currently no time-based "
                                 "events are used")

            for idx, (t, x) in enumerate(zip(res.t[::-1], res.x[::-1, :])):
                state_rate = prob.state_equation_function(t, x)

                if (idx == last_res_idx) and (prob is not self.sim_problems[0]):
                    next_prob = self.sim_problems[self.sim_problems.index(prob)-1]

                    f_plus = np.zeros(next_prob.dim_state)
                    plus_rate = state_rate

                    # NOTE / TODO: should enforce that all states in all ODEs exist
                    # in eachother (even if only as output). Don't like assuming
                    # zero
                    # state_update = np.zeros(next_prob.dim_state)
                    state_update = np.ones(next_prob.dim_state)*np.inf
                    dh_dx = np.zeros((next_prob.dim_state,)*2)
                    dh_dparam = np.zeros((next_prob.dim_state, len(param_dict)))

                    # here and co-state assume number of states is only decreasing
                    # forward in time
                    for state_name in next_prob.states.keys():
                        state_idx = list(next_prob.states.keys()).index(state_name)

                        if state_name in prob.states.keys():
                            f_plus[
                                state_idx
                            ] = plus_rate[list(prob.states.keys()).index(state_name)]

                            # state_update[
                            #    list(next_prob.states.keys()).index(state_name)
                            # ] = x[list(prob.states.keys()).index(state_name)]

                            # TODO: make sure index multiplying next_pronb costate
                            # lines up -- since costate is pre-filled to next_prob's
                            # order, the continuous terms should be right
                            # column should map to
                            dh_dx[state_idx, state_idx] = 1.

                        elif state_name in prob.outputs.keys():
                            state_update[
                                state_idx
                            ] = res.y[-1, list(prob.outputs.keys()).index(state_name)]

                            dh_j_dx = prob.compute_totals(
                                [state_name],
                                prob.states.keys(),
                                return_format='array').squeeze()

                            dh_dparam[state_idx, :] = prob.compute_totals(
                                [state_name],
                                list(param_dict.keys()),
                                return_format='array'
                            ).squeeze()

                            for state_name_2 in prob.states.keys():
                                # I'm actually computing dh_dx.T
                                # dh_dx rows are new state, columns are old state
                                # now, dh_dx.T rows are old state, columns are new
                                # so I think this is right
                                dh_dx[
                                    list(next_prob.states.keys()).index(state_name_2),
                                    state_idx,
                                ] = dh_j_dx[list(prob.states.keys()).index(state_name_2)]

                        else:
                            state_update[
                                state_idx
                            ] = 0.

                    f_pluses.append(f_plus)
                    state_updates.append(state_update)
                    dh_dxs.append(dh_dx)
                    dh_dparams.append(dh_dparam)

                state_rate_names = [val['rate'] for _, val in prob.states.items()]
                df_dx_data[idx, :, :] = prob.compute_totals(state_rate_names,
                                                            prob.states.keys(),
                                                            return_format='array').T
                if param_dict:
                    df_dparam_data[idx, ...] = prob.compute_totals(
                        state_rate_names,
                        list(param_dict.keys()),
                        return_format='array'
                    )

            k = min(3, res.t.shape[0]-1)
            skip_interp = (k == 1) and np.isclose(res.t[0], res.t[1])

            # TODO: why is this failing?
            if skip_interp:
                mean_df_dx = np.mean(df_dx_data, axis=0)

                df_dxs.append(lambda t: mean_df_dx)
            else:
                try:
                    df_dxs.append(interpolate.make_interp_spline(
                        tf_total - res.t[::-1],
                        df_dx_data,
                        k=k
                    ))
                except Exception as e:

                    print("EXCEPTION!", k, res.t, df_dx_data, sep='\n')
                    print("check: ", (k == 1) and (res.t[0] == res.t[1]))
                    breakpoint()
                    raise e

            if param_dict:
                if skip_interp:
                    mean_df_dparam = np.mean(df_dparam_data, axis=0)
                    df_dparams.append(lambda t: mean_df_dparam)
                else:
                    df_dparams.append(interpolate.make_interp_spline(
                        tf_total - res.t[::-1],
                        df_dparam_data,
                        k=k
                    ))
            else:
                df_dparams.append(None)

        if self.DEBUG:
            print("data....")
            print("dgs", dg_dxs)
            print("f-", f_minuses)
            print("f+", f_pluses)

            print("size check:", len(self.sim_problems), len(dg_dxs), len(f_minuses),
                  len(f_pluses), )

        # main loop
        for output, costate_ic, param_deriv in zip(self.all_traj_outputs, costate_ics,
                                                   param_derivs):

            output_name = self.all_traj_outputs[output]["name"]
            next_prob = self.sim_problems[-1]
            costate = costate_ic
            lamda_dot_plus = np.zeros_like(costate)

            # self.sim_results[-1].x[-1, next_prob.state_names.index(output)]
            if self.DEBUG:
                print("\nstarting partial for %s" % output, costate)

            dg_dt = 0.

            for (
                phase_idx,
                res,
                prob,
                df_dx,
                df_dparam,
                dg_dx,
                f_minus,
                f_plus,
                state_update,
                dh_dx,
                dh_dparam,
            ) in zip(
                range(len(self.sim_results), 0, -1),
                self.sim_results[::-1],
                self.sim_problems[::-1],
                df_dxs,
                df_dparams,
                dg_dxs,
                f_minuses,
                f_pluses,
                state_updates,
                dh_dxs,
                dh_dparams,
            ):

                t0, tf = tf_total - res.t[[-1, 0]]

                # assumes only 1 of time, state, or output dependence
                # assume no discontinuous state update, would need an API for that in
                # compute as well --
                # but assume some form of event has happened

                # already checked that event_channel_names was well-defined in the
                # pre-compute, so will just assign the co-state just once
                for channel_idx, channel_name in enumerate(prob.event_channel_names):
                    if np.argmin(np.abs(res.e[-1, :])) not in [channel_idx]:
                        continue

                    state_disc = res.x[-1] - state_update
                    state_disc[np.where(np.isinf(state_update))] = 0.

                    if channel_name != prob.t_name:
                        lamda_dot = df_dx(res.t[-1]) @ costate
                        # lamda_dot_plus = lamda_dot
                        if self.DEBUG:
                            if np.any(state_disc):
                                print("update is non-zero!", prob, prob.states.keys(),
                                      state_disc, costate, lamda_dot)
                                print(
                                    "inner product becomes...",
                                    state_disc[None,
                                               :] @ dh_dx @ lamda_dot_plus[:, None],
                                    state_disc[None,
                                               :] @ dh_dx.T @ lamda_dot_plus[:, None]
                                )
                            print("dh_dx for", prob, prob.states.keys(), "\n",  dh_dx)
                            print("costate", costate)
                        costate_update_terms = [
                            dh_dx.T @ costate[:, None],
                            # costate[:, None],
                            # TODO: should this be f_plus? probably not
                            (dg_dx.T @ (f_plus - f_minus)
                             [None, :] @ costate[:, None]) / (dg_dx@f_minus),
                            # don't believe in lamda_dot terms anymore
                            # -(dg_dx.T @ state_disc[None, :] @ dh_dx.T @ lamda_dot_plus[:, None]) / (dg_dx@f_minus),

                        ]

                        # TODO: is this wrong?
                        costate[:] = np.sum(costate_update_terms, axis=0).squeeze()

                    if (
                        (event_key := (prob, channel_name, channel_idx))
                        in self.traj_event_trigger_input
                    ):
                        event_trigger_name = self.traj_event_trigger_input[event_key]["name"]
                        if self.DEBUG:
                            print("setting event trigger data", event_trigger_name)
                        J[output_name, event_trigger_name] = (
                            + costate[None, :] @ (f_minus - f_plus) /
                            (dg_dt + dg_dx@f_minus)
                            # +(lamda_dot_plus[None, :] @ dh_dx @ state_disc[None, :])/(dg_dt + dg_dx@f_minus)
                        )

                    # how to account for terminal event? through costate IC.
                    # TODO: Is this wrong?
                    param_deriv += (costate[None, :] @ dh_dparam).squeeze()

                # build co-state systems

                def co_state_rate(t, costate, *args):
                    return df_dx(t) @ costate

                if self.DEBUG:
                    print('dim_state:', prob.dim_state, "ic:", costate)

                costate_sys = DynamicalSystem(state_equation_function=co_state_rate,
                                              dim_state=prob.dim_state)
                costate_sys.initial_condition = costate

                # simulate co-state system
                co_res = costate_sys.simulate(
                    (t0, tf), integrator_options=self.adjoint_int_opts)
                costate_reses[output].append(co_res)

                if param_dict:
                    df_dparam_val = df_dparam(co_res.t)
                    param_deriv_integrand_data = np.matmul(
                        co_res.x[:, None, :],
                        df_dparam_val
                    ).squeeze()
                    try:
                        param_deriv_integrand = interpolate.make_interp_spline(
                            co_res.t,
                            np.atleast_1d(param_deriv_integrand_data),
                            # k=df_dparam.k
                            k=min(3, co_res.t.shape[0]-1)
                        )
                    except ValueError as e:
                        print(
                            "HIT VALUE ERROR!",
                            output,
                            prob,
                            co_res.t.shape,
                            co_res.x.shape,
                            df_dparam_val.shape,
                            df_dparam.k,
                            "final_results:\n\n",
                            t0, tf,
                            co_res.t,
                            co_res.x,
                        )
                        raise e
                    param_deriv_integrand_antideriv = param_deriv_integrand.antiderivative()

                    # TODO: is the sign wrong here?
                    param_deriv -= (
                        param_deriv_integrand_antideriv(t0)
                        - param_deriv_integrand_antideriv(tf)
                    )

                # consume initial condition
                if prob is not self.sim_problems[0]:
                    next_prob = self.sim_problems[self.sim_problems.index(prob)-1]
                else:
                    break
                costate = np.zeros(next_prob.dim_state)
                lamda_dot_plus = np.zeros_like(costate)
                lamda_dot_plus_rate = co_state_rate(co_res.t[-1], co_res.x[-1])

                # TODO: do co-states need unit changes? probably not...
                for state_name in prob.state_names:
                    costate[list(next_prob.states.keys()).index(
                        state_name)] = co_res.x[-1, list(prob.states.keys()).index(state_name)]
                    lamda_dot_plus[
                        list(next_prob.states.keys()).index(state_name)
                    ] = lamda_dot_plus_rate[list(prob.states.keys()).index(state_name)]

            for state_to_deriv, metadata in self.traj_initial_state_input.items():
                param_name = metadata["name"]
                J[output_name, param_name] = costate_reses[output][-1].x[
                    -1,
                    list(prob.states.keys()).index(state_to_deriv)
                ]
            for param_deriv_val, param_deriv_name in zip(param_deriv, param_dict):
                J[output_name, param_deriv_name] = param_deriv_val
        self.costate_reses = costate_reses



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.hydraulics import (AltHydraulicsGroupMass,
                                                           TransportHydraulicsGroupMass)
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (Version,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission

_sys_press = 'aircraft:hydraulics:dimensions:sys_press'
_wing_ref_area = 'aircraft:wing:dimensions:area'
_fuse_planform_area = 'TBD:fuselage_planform_area'
_horiz_tail_wetted_area = 'aircraft:horizontal_tail:dimensions:wetted_area'
_horiz_tail_thick_chord = 'aircraft:horizontal_tail:dimensions:thickness_to_chord_ratio'
_vert_tail_area = 'aircraft:vertical_tails:dimensions:area'
_wing_engine_count_factor = 'aircraft:propulsion:control:wing_engine_count_factor'
_fuse_engine_count_factor = 'aircraft:propulsion:control:fuselage_engine_count_factor'
_var_sweep_mass_penalty = 'aircraft:wing:mass:var_sweep_mass_penalty'
_max_mach = 'aircraft:design:dimensions:max_mach'
_hydraulics_group_mass = 'TBD:hydraulics_mass'


class TransportHydraulicsGroupMassTest(unittest.TestCase):
    '''
    Tests transport/GA hydraulics mass calculation.
    '''

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'hydraulics',
            TransportHydraulicsGroupMass(aviary_options=get_flops_inputs(case_name)),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Fuselage.PLANFORM_AREA,
                        Aircraft.Hydraulics.SYSTEM_PRESSURE,
                        Aircraft.Hydraulics.MASS_SCALER,
                        Aircraft.Wing.AREA,
                        Aircraft.Wing.VAR_SWEEP_MASS_PENALTY],
            output_keys=Aircraft.Hydraulics.MASS,
            version=Version.TRANSPORT,
            tol=4.0e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class AltHydraulicsGroupMassTest(unittest.TestCase):
    '''
    Tests alternate hydraulics mass calculation.
    '''

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'hydraulics',
            AltHydraulicsGroupMass(aviary_options=get_flops_inputs(case_name)),
            promotes_outputs=['*'],
            promotes_inputs=['*']
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Wing.AREA,
                        Aircraft.HorizontalTail.WETTED_AREA,
                        Aircraft.HorizontalTail.THICKNESS_TO_CHORD,
                        Aircraft.VerticalTail.AREA,
                        Aircraft.Hydraulics.MASS_SCALER],
            output_keys=Aircraft.Hydraulics.MASS,
            version=Version.ALTERNATE)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == '__main__':
    unittest.main()



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.wing_simple import SimpleWingBendingFact
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft


class SimpleWingBendingFactTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    # Only dataset that uses the simple wing.
    @parameterized.expand(get_flops_case_names(only=['LargeSingleAisle2FLOPS', 'LargeSingleAisle2FLOPSalt']),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            "wing",
            SimpleWingBendingFact(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.Wing.AREA,
                        Aircraft.Wing.SPAN,
                        Aircraft.Wing.TAPER_RATIO,
                        Aircraft.Wing.THICKNESS_TO_CHORD,
                        Aircraft.Wing.STRUT_BRACING_FACTOR,
                        Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR,
                        Aircraft.Wing.ASPECT_RATIO,
                        Aircraft.Wing.SWEEP],
            output_keys=[Aircraft.Wing.BENDING_FACTOR,
                         Aircraft.Wing.ENG_POD_INERTIA_FACTOR],
            atol=1e-11,
            rtol=1e-11)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



"""
Utilities used for testing Aviary code.
"""
from dymos.utils.testing_utils import assert_timeseries_near_equal


def warn_timeseries_near_equal(t_ref, x_ref, t_check, x_check, abs_tolerance=None,
                               rel_tolerance=None):
    """
    Wraps timeseries check with a try block that prints a warning if the assert fails.
    """
    try:
        assert_timeseries_near_equal(t_ref, x_ref, t_check, x_check, abs_tolerance=abs_tolerance,
                                     rel_tolerance=rel_tolerance)
    except AssertionError as exc:
        print("Warning:\n", str(exc))
